# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import lldb
import struct

# this is an LLDB pretty printer for libtorrent types. To use in LLDB run:
#
#   command script import tools/libtorrent_lldb.py
#
# or add it to your ~/.lldbinit

def __lldb_init_module (debugger, dict):
    debugger.HandleCommand("type summary add -x \"^libtorrent::digest32<.+>$\" -F " +
        "libtorrent_lldb.print_hash -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^libtorrent::sha256_hash$\" -F " +
        "libtorrent_lldb.print_hash -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^libtorrent::sha1_hash$\" -F " +
        "libtorrent_lldb.print_hash -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^libtorrent::span<.+>$\" -F " +
        "libtorrent_lldb.print_span -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^libtorrent::flags::bitfield_flag<.+>$\" -F " +
        "libtorrent_lldb.print_flag -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^boost::asio::ip::basic_endpoint<.+>$\" -F " +
        "libtorrent_lldb.print_endpoint -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^libtorrent::bitfield$\" -F " +
        "libtorrent_lldb.print_bitfield -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^libtorrent::typed_bitfield<.+>$\" -F " +
        "libtorrent_lldb.print_bitfield -p -w libtorrent")
    debugger.HandleCommand("type summary add -x \"^libtorrent::aux::strong_typedef<.+>$\" -F " +
        "libtorrent_lldb.print_strong_type -p -w libtorrent")
    debugger.HandleCommand("type category enable libtorrent")

def print_hash(valobj, internal_dict):

    if valobj.GetType().IsReferenceType():
        valobj = valobj.Dereference()

    data = valobj.GetChildMemberWithName("m_number").GetData().uint8s
    return bytes(data).hex()

def print_flag(valobj, internal_dict):

    if valobj.GetType().IsReferenceType():
        valobj = valobj.Dereference()

    data = valobj.GetChildMemberWithName("m_val").GetValueAsUnsigned()
    return "({}) {:b}".format(valobj.GetType().name, data)

def swap16(i):
    return struct.unpack("<H", struct.pack(">H", i))[0]

def pairs(lst):
    for i in range(0, len(lst), 2):
        yield lst[i:i+2]

def print_endpoint(valobj, internal_dict):

    if valobj.GetType().IsReferenceType():
        valobj = valobj.Dereference()

    union = valobj.GetChildMemberWithName("impl_").GetChildMemberWithName("data_")
    family = union.GetChildMemberWithName("base").GetChildMemberWithName("sa_family").GetValueAsUnsigned()

    if family == 2:
        a = union.GetChildMemberWithName("v4").GetChildMemberWithName("sin_addr").GetData().uint8s
        addr = ".".join([f"{b}" for b in a ])
        p = swap16(union.GetChildMemberWithName("v4").GetChildMemberWithName("sin_port").GetValueAsUnsigned())
        return "{}:{}".format(addr, p)
    else:
        a = union.GetChildMemberWithName("v6").GetChildMemberWithName("sin6_addr").GetData().uint8s
        p = swap16(union.GetChildMemberWithName("v6").GetChildMemberWithName("sin6_port").GetValueAsUnsigned())
        addr = ":".join(x+y for x, y in pairs(["{:02x}".format(b) for b in a]))
        return "[{}]:{}".format(addr, p)

def print_bitfield(valobj, internal_dict):

    if valobj.GetType().IsReferenceType():
        valobj = valobj.Dereference()

    array = valobj.GetChildMemberWithName("m_buf").GetChildMemberWithName("__ptr_").GetChildMemberWithName("__value_")
    size = array.Dereference().GetValueAsUnsigned()
    ret = "size: {} bits | ".format(size)
    for idx in range((size + 31) // 32):
        item = array.GetChildAtIndex(idx + 1, lldb.eNoDynamicValues, True)
        buffer = item.GetData().uint8s
        for b in buffer:
            ret += "{:08b}".format(int(b))
            size -= 8

    return ret

def print_span(valobj, internal_dict):

    if valobj.GetType().IsReferenceType():
        valobj = valobj.Dereference()

    array = valobj.GetChildMemberWithName("m_ptr")
    size = valobj.GetChildMemberWithName("m_len").GetValueAsSigned()
    ret = "size = {}".format(size)
    for idx in range(size):
        if idx == 0:
            item = array.Dereference()
        else:
            item = array.GetChildAtIndex(idx, lldb.eNoDynamicValues, True)
        ret += "\n[{}] = {}".format(idx, item.summary)
    return ret


def print_strong_type(valobj, internal_dict):

    if valobj.GetType().IsReferenceType():
        valobj = valobj.Dereference()

    name = valobj.GetType().name
    data = valobj.GetChildMemberWithName("m_val").GetValue()
    if "piece_index_tag" in name:
        name = "piece_index"
    elif "file_index_tag" in name:
        name = "file_index"
    elif "queue_position_tag" in name:
        name = "queue_pos"
    elif "piece_extent_tag" in name:
        name = "piece_extent"
    elif "storage_index_tag_t" in name:
        name = "storage_index"
    elif "disconnect_severity_tag" in name:
        name = "disconnect_severity"
        val = valobj.GetChildMemberWithName("m_val").GetValueAsUnsigned()
        if val == 0:
            data = "normal"
        elif val == 1:
            data = "failure"
        elif val == 2:
            data = "peer_error"
        else:
            data = "<unknown> ({})".format(val)
    elif "prio_index_tag_t" in name:
        name = "prio_index"
    elif "port_mapping_tag" in name:
        name = "port_mapping"
    elif "dl_queue_tag" in name or name == "libtorrent::download_queue_t":
        name = "download_queue"
        val = valobj.GetChildMemberWithName("m_val").GetValueAsUnsigned()
        if val == 0:
            data = "piece_downloading";
        elif val == 1:
            data = "piece_full";
        elif val == 2:
            data = "piece_finished";
        elif val == 3:
            data = "piece_zero_prio";
        elif val == 4:
            data = "piece_open";
        elif val == 5:
            data = "piece_downloading_reverse";
        elif val == 6:
            data = "piece_full_reverse";
        else:
            data = "<unknown> ({})".format(val)
    elif "piece_extent_tag" in name:
        name = "piece_extent"
    elif "picker_options_tag" in name:
        name = "picker_options"
        val = valobj.GetChildMemberWithName("m_val").GetValueAsUnsigned()
        flags = []
        if (val & 1) != 0:
            flags.append("rarest_first")
        if (val & 2) != 0:
            flags.append("reverse")
        if (val & 4) != 0:
            flags.append("on_parole")
        if (val & 8) != 0:
            flags.append("prioritize_partials")
        if (val & 16) != 0:
            flags.append("sequential")
        if (val & 64) != 0:
            flags.append("align_expanded_pieces")
        if (val & 128) != 0:
            flags.append("piece_extent_affinity")
        data = "|".join(flags)
    else:
        name = ""

    return "({}) {}".format(name, data)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
from __future__ import print_function

import sys


def indent(line):
    if line == '':
        return None
    end = 0
    for c in line:
        end += 1
        if " \t" not in c:
            return line[:end]
    return line


start_block = False
filter_indent = None

for line in open(sys.argv[1]):

    if line == '\n':
        continue

    if filter_indent:
        if line.startswith(filter_indent):
            continue
        else:
            filter_indent = None

    if line.strip().startswith('.. '):
        start_block = True
        continue

    if line.endswith('::\n'):
        start_block = True
        continue

    if start_block:
        filter_indent = indent(line)
        start_block = False
        continue

    sys.stdout.write(line)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4


import libtorrent as lt

import unittest
import time
import datetime
import os
import shutil
import binascii
import subprocess as sub
import sys
import pickle
import threading
import tempfile
import socket
import select
import logging
import ssl
import http.server
import functools

import dummy_data

# include terminal interface for travis parallel executions of scripts which use
# terminal features: fix multiple stdin assignment at termios.tcgetattr
if os.name != 'nt':
    import pty

settings = {
    'alert_mask': lt.alert.category_t.all_categories,
    'enable_dht': False, 'enable_lsd': False, 'enable_natpmp': False,
    'enable_upnp': False, 'listen_interfaces': '0.0.0.0:0', 'file_pool_size': 1}


def has_deprecated():
    return hasattr(lt, 'version')


class test_create_torrent(unittest.TestCase):

    def test_from_torrent_info(self):
        ti = lt.torrent_info('unordered.torrent')
        print(ti.ssl_cert())
        ct = lt.create_torrent(ti)
        entry = ct.generate()
        content = lt.bencode(entry).strip()
        with open('unordered.torrent', 'rb') as f:
            file_content = bytearray(f.read().strip())
            print(content)
            print(file_content)
            print(entry)
            self.assertEqual(content, file_content)

    def test_from_scratch(self):
        fs = lt.file_storage()
        fs.add_file('test/file1', 1000)
        fs.add_file('test/file2', 2000)
        self.assertEqual(fs.file_name(0), 'file1')
        self.assertEqual(fs.file_name(1), 'file2')
        ct = lt.create_torrent(fs, 0, lt.create_torrent.canonical_files_no_tail_padding)
        ct.add_url_seed('foo')
        ct.add_http_seed('bar')
        ct.add_tracker('bar')
        ct.set_root_cert('1234567890')
        ct.add_collection('1337')
        for i in range(ct.num_pieces()):
            ct.set_hash(i, b'abababababababababab')
        entry = ct.generate()
        encoded = lt.bencode(entry)
        print(encoded)

        # zero out the creation date:
        encoded = encoded.split(b'13:creation datei', 1)
        encoded[1] = b'0e' + encoded[1].split(b'e', 1)[1]
        encoded = b'13:creation datei'.join(encoded)

        self.assertEqual(encoded, b'd8:announce3:bar13:creation datei0e9:httpseeds3:bar4:infod11:collectionsl4:1337e5:filesld6:lengthi1000e4:pathl5:file1eed4:attr1:p6:lengthi15384e4:pathl4:.pad5:15384eed6:lengthi2000e4:pathl5:file2eee4:name4:test12:piece lengthi16384e6:pieces40:abababababababababababababababababababab8:ssl-cert10:1234567890e8:url-list3:fooe')


class test_session_stats(unittest.TestCase):

    def test_add_torrent_params(self):
        atp = lt.add_torrent_params()

        for field_name in dir(atp):
            field = getattr(atp, field_name)
            print(field_name, field)

        atp.renamed_files = {}
        atp.merkle_tree = []
        atp.unfinished_pieces = {}
        atp.have_pieces = []
        atp.banned_peers = []
        atp.verified_pieces = []
        atp.piece_priorities = []
        atp.url_seeds = []

    def test_unique(self):
        metrics = lt.session_stats_metrics()
        self.assertTrue(len(metrics) > 40)
        idx = set()
        for m in metrics:
            self.assertTrue(m.value_index not in idx)
            idx.add(m.value_index)

    def test_find_idx(self):
        self.assertEqual(lt.find_metric_idx("peer.error_peers"), 0)


class test_torrent_handle(unittest.TestCase):

    def setup(self):
        self.ses = lt.session(settings)
        self.ti = lt.torrent_info('url_seed_multi.torrent')
        self.h = self.ses.add_torrent({
            'ti': self.ti, 'save_path': os.getcwd(),
            'flags': lt.torrent_flags.default_flags})

    def test_add_torrent_error(self):
        self.ses = lt.session(settings)
        self.ti = lt.torrent_info('url_seed_multi.torrent')
        with self.assertRaises(RuntimeError):
            self.ses.add_torrent({'ti': self.ti, 'save_path': os.getcwd(), 'info_hashes': b'abababababababababab'})

    def test_move_storage(self):
        self.setup()
        self.h.move_storage(u'test-dir')
        self.h.move_storage(b'test-dir2')
        self.h.move_storage('test-dir3')
        self.h.move_storage(u'test-dir', flags=lt.move_flags_t.dont_replace)
        self.h.move_storage(u'test-dir', flags=2)
        self.h.move_storage(b'test-dir2', flags=2)
        self.h.move_storage('test-dir3', flags=2)

    def test_torrent_handle(self):
        self.setup()
        self.assertEqual(self.h.get_file_priorities(), [4, 4])
        self.assertEqual(self.h.get_piece_priorities(), [4])

        self.h.prioritize_files([0, 1])
        # workaround for asynchronous priority update
        time.sleep(1)
        self.assertEqual(self.h.get_file_priorities(), [0, 1])

        self.h.prioritize_pieces([0])
        self.assertEqual(self.h.get_piece_priorities(), [0])

        # also test the overload that takes a list of piece->priority mappings
        self.h.prioritize_pieces([(0, 1)])
        self.assertEqual(self.h.get_piece_priorities(), [1])
        self.h.connect_peer(('127.0.0.1', 6881))
        self.h.connect_peer(('127.0.0.2', 6881), source=4)
        self.h.connect_peer(('127.0.0.3', 6881), flags=2)
        self.h.connect_peer(('127.0.0.4', 6881), flags=2, source=4)

        torrent_files = self.h.torrent_file()
        print(torrent_files.map_file(0, 0, 0).piece)

        print(self.h.queue_position())

    def test_torrent_handle_in_set(self):
        self.setup()
        torrents = set()
        torrents.add(self.h)

        # get another instance of a torrent_handle that represents the same
        # torrent. Make sure that when we add it to a set, it just replaces the
        # existing object
        t = self.ses.get_torrents()
        self.assertEqual(len(t), 1)
        for h in t:
            torrents.add(h)

        self.assertEqual(len(torrents), 1)

    def test_torrent_handle_in_dict(self):
        self.setup()
        torrents = {}
        torrents[self.h] = 'foo'

        # get another instance of a torrent_handle that represents the same
        # torrent. Make sure that when we add it to a dict, it just replaces the
        # existing object
        t = self.ses.get_torrents()
        self.assertEqual(len(t), 1)
        for h in t:
            torrents[h] = 'bar'

        self.assertEqual(len(torrents), 1)
        self.assertEqual(torrents[self.h], 'bar')

    def test_replace_trackers(self):
        self.setup()
        trackers = []
        for idx, tracker_url in enumerate(('udp://tracker1.com', 'udp://tracker2.com')):
            tracker = lt.announce_entry(tracker_url)
            tracker.tier = idx
            tracker.fail_limit = 2
            trackers.append(tracker)
            self.assertEqual(tracker.url, tracker_url)
        self.h.replace_trackers(trackers)
        new_trackers = self.h.trackers()
        self.assertEqual(new_trackers[0]['url'], 'udp://tracker1.com')
        self.assertEqual(new_trackers[1]['tier'], 1)
        self.assertEqual(new_trackers[1]['fail_limit'], 2)

    def test_pickle_trackers(self):
        """Test lt objects converters are working and trackers can be pickled"""
        self.setup()
        tracker = lt.announce_entry('udp://tracker1.com')
        tracker.tier = 0
        tracker.fail_limit = 1
        trackers = [tracker]
        self.h.replace_trackers(trackers)
        # wait a bit until the endpoints list gets populated
        while len(self.h.trackers()[0]['endpoints']) == 0:
            time.sleep(0.1)

        trackers = self.h.trackers()
        self.assertEqual(trackers[0]['url'], 'udp://tracker1.com')
        # this is not necessarily 0, it could also be (EHOSTUNREACH) if the
        # local machine doesn't support the address family
        expect_value = trackers[0]['endpoints'][0]['info_hashes'][0]['last_error']['value']
        pickled_trackers = pickle.dumps(trackers)
        unpickled_trackers = pickle.loads(pickled_trackers)
        self.assertEqual(unpickled_trackers[0]['url'], 'udp://tracker1.com')
        self.assertEqual(unpickled_trackers[0]['endpoints'][0]['info_hashes'][0]['last_error']['value'], expect_value)

    def test_file_status(self):
        self.setup()
        status = self.h.file_status()
        print(status)

    def test_piece_deadlines(self):
        self.setup()
        self.h.clear_piece_deadlines()

    def test_status_last_uploaded_dowloaded(self):
        # we want to check at seconds precision but can't control session
        # time, wait for next full second to prevent second increment
        time.sleep(1 - datetime.datetime.now().microsecond / 1000000.0)

        self.setup()
        st = self.h.status()
        for attr in dir(st):
            print('%s: %s' % (attr, getattr(st, attr)))
        # last upload and download times are at session start time
        self.assertEqual(st.last_upload, None)
        self.assertEqual(st.last_download, None)

    def test_serialize_trackers(self):
        """Test to ensure the dict contains only python built-in types"""
        self.setup()
        self.h.add_tracker({'url': 'udp://tracker1.com'})
        tr = self.h.trackers()[0]
        # wait a bit until the endpoints list gets populated
        while len(tr['endpoints']) == 0:
            time.sleep(0.1)
            tr = self.h.trackers()[0]
        import json
        print(json.dumps(self.h.trackers()[0]))

    def test_torrent_status(self):
        self.setup()
        st = self.h.status()
        ti = st.handle
        self.assertEqual(ti.info_hashes(), self.ti.info_hashes())
        # make sure we can compare torrent_status objects
        st2 = self.h.status()
        self.assertEqual(st2, st)
        print(st2)

    def test_read_resume_data(self):

        resume_data = lt.bencode({
            'file-format': 'libtorrent resume file',
            'info-hash': 'abababababababababab',
            'name': 'test',
            'save_path': '.',
            'peers': '\x01\x01\x01\x01\x00\x01\x02\x02\x02\x02\x00\x02',
            'file_priority': [0, 1, 1]})
        tp = lt.read_resume_data(resume_data)

        self.assertEqual(tp.name, 'test')
        self.assertEqual(tp.info_hashes.v1, lt.sha1_hash('abababababababababab'))
        self.assertEqual(tp.file_priorities, [0, 1, 1])
        self.assertEqual(tp.peers, [('1.1.1.1', 1), ('2.2.2.2', 2)])

        ses = lt.session(settings)
        h = ses.add_torrent(tp)
        for attr in dir(tp):
            print('%s: %s' % (attr, getattr(tp, attr)))

        h.connect_peer(('3.3.3.3', 3))

        for i in range(0, 10):
            alerts = ses.pop_alerts()
            for a in alerts:
                print(a.message())
            time.sleep(0.1)

    def test_scrape(self):
        self.setup()
        # this is just to make sure this function can be called like this
        # from python
        self.h.scrape_tracker()

    def test_unknown_torrent_parameter(self):
        self.ses = lt.session(settings)
        try:
            self.h = self.ses.add_torrent({'unexpected-key-name': ''})
            self.assertFalse('should have thrown an exception')
        except KeyError as e:
            print(e)

    def test_torrent_parameter(self):
        self.ses = lt.session(settings)
        self.ti = lt.torrent_info('url_seed_multi.torrent')
        self.h = self.ses.add_torrent({
            'ti': self.ti,
            'save_path': os.getcwd(),
            'trackers': ['http://test.com/announce'],
            'dht_nodes': [('1.2.3.4', 6881), ('4.3.2.1', 6881)],
            'file_priorities': [1, 1],
            'http_seeds': ['http://test.com/file3'],
            'url_seeds': ['http://test.com/announce-url'],
            'peers': [('5.6.7.8', 6881)],
            'banned_peers': [('8.7.6.5', 6881)],
            'renamed_files': {0: 'test.txt', 2: 'test.txt'}
        })
        self.st = self.h.status()
        self.assertEqual(self.st.save_path, os.getcwd())
        trackers = self.h.trackers()
        self.assertEqual(len(trackers), 1)
        self.assertEqual(trackers[0].get('url'), 'http://test.com/announce')
        self.assertEqual(trackers[0].get('tier'), 0)
        self.assertEqual(self.h.get_file_priorities(), [1, 1])
        self.assertEqual(self.h.http_seeds(), ['http://test.com/file3'])
        # url_seeds was already set, test that it did not get overwritten
        self.assertEqual(self.h.url_seeds(),
                         ['http://test.com/announce-url/', 'http://test.com/file/'])
        # piece priorities weren't set explicitly, but they were updated by the
        # file priorities being set
        self.assertEqual(self.h.get_piece_priorities(), [1])
        self.assertEqual(self.st.verified_pieces, [])


class TestAddPiece(unittest.TestCase):

    def setUp(self):
        self.dir = tempfile.TemporaryDirectory()
        self.session = lt.session(settings)
        self.ti = lt.torrent_info(dummy_data.DICT)
        self.atp = lt.add_torrent_params()
        self.atp.ti = self.ti
        self.atp.save_path = self.dir.name
        self.handle = self.session.add_torrent(self.atp)
        self.wait_for(lambda: self.handle.status().state != lt.torrent_status.checking_files
                      and self.handle.status().state != lt.torrent_status.checking_resume_data, msg="checking")

    def wait_for(self, condition, msg="condition", timeout=5):
        deadline = time.time() + timeout
        while not condition():
            self.assertLess(time.time(), deadline, msg="%s timed out" % msg)
            time.sleep(0.1)

    def wait_until_torrent_finished(self):
        self.wait_for(lambda: self.handle.status().progress == 1.0, msg="progress")

        def file_written():
            with open(os.path.join(self.dir.name.encode(), dummy_data.NAME), mode="rb") as f:
                return f.read() == dummy_data.DATA

        self.wait_for(file_written, msg="file write")

    def test_with_str(self):
        for i, data in enumerate(dummy_data.PIECES):
            self.handle.add_piece(i, data.decode(), 0)

        self.wait_until_torrent_finished()

    def test_with_bytes(self):
        for i, data in enumerate(dummy_data.PIECES):
            self.handle.add_piece(i, data, 0)

        self.wait_until_torrent_finished()


class test_load_torrent(unittest.TestCase):

    def test_bytearray(self):
        # a bytearray object is interpreted as a bencoded buffer
        atp = lt.load_torrent_buffer(bytearray(lt.bencode({'info': {
            'name': 'test_torrent', 'length': 1234,
            'piece length': 16 * 1024,
            'pieces': 'aaaaaaaaaaaaaaaaaaaa'}})))
        self.assertEqual(atp.ti.num_files(), 1)

    def test_bytes(self):
        # a bytes object is interpreted as a bencoded buffer
        atp = lt.load_torrent_buffer(bytes(lt.bencode({'info': {
            'name': 'test_torrent', 'length': 1234,
            'piece length': 16 * 1024,
            'pieces': 'aaaaaaaaaaaaaaaaaaaa'}})))
        self.assertEqual(atp.ti.num_files(), 1)

    def test_info_section(self):
        atp = lt.load_torrent_file('base.torrent')

        self.assertTrue(len(atp.ti.info_section()) != 0)
        self.assertTrue(len(atp.ti.hash_for_piece(0)) != 0)

class test_torrent_info(unittest.TestCase):

    def test_non_ascii_file(self):
        try:
            shutil.copy('base.torrent', 'base-\u745E\u5177.torrent')
        except shutil.SameFileError:
            pass
        ti = lt.torrent_info('base-\u745E\u5177.torrent')

        self.assertTrue(len(ti.info_section()) != 0)
        self.assertTrue(len(ti.hash_for_piece(0)) != 0)

    def test_bencoded_constructor(self):
        # things that can be converted to a bencoded entry, will be interpreted
        # as such and encoded
        info = lt.torrent_info({'info': {
            'name': 'test_torrent', 'length': 1234,
            'piece length': 16 * 1024,
            'pieces': 'aaaaaaaaaaaaaaaaaaaa'}})

        self.assertEqual(info.num_files(), 1)

        f = info.files()
        self.assertEqual(f.file_path(0), 'test_torrent')
        self.assertEqual(f.file_name(0), 'test_torrent')
        self.assertEqual(f.file_size(0), 1234)
        self.assertEqual(info.total_size(), 1234)
        self.assertEqual(info.creation_date(), 0)

    def test_bytearray(self):
        # a bytearray object is interpreted as a bencoded buffer
        info = lt.torrent_info(bytearray(lt.bencode({'info': {
            'name': 'test_torrent', 'length': 1234,
            'piece length': 16 * 1024,
            'pieces': 'aaaaaaaaaaaaaaaaaaaa'}})))
        self.assertEqual(info.num_files(), 1)

    def test_bytes(self):
        # a bytes object is interpreted as a bencoded buffer
        info = lt.torrent_info(bytes(lt.bencode({'info': {
            'name': 'test_torrent', 'length': 1234,
            'piece length': 16 * 1024,
            'pieces': 'aaaaaaaaaaaaaaaaaaaa'}})))
        self.assertEqual(info.num_files(), 1)

    def test_load_decode_depth_limit(self):
        self.assertRaises(RuntimeError, lambda: lt.torrent_info(
            {'test': {'test': {'test': {'test': {'test': {}}}}}, 'info': {
                'name': 'test_torrent', 'length': 1234,
                'piece length': 16 * 1024,
                'pieces': 'aaaaaaaaaaaaaaaaaaaa'}}, {'max_decode_depth': 1}))

    def test_load_max_pieces_limit(self):
        self.assertRaises(RuntimeError, lambda: lt.torrent_info(
            {'info': {
                'name': 'test_torrent', 'length': 1234000,
                'piece length': 16 * 1024,
                'pieces': 'aaaaaaaaaaaaaaaaaaaa'}}, {'max_pieces': 1}))

    def test_load_max_buffer_size_limit(self):
        self.assertRaises(RuntimeError, lambda: lt.torrent_info(
            {'info': {
                'name': 'test_torrent', 'length': 1234000,
                'piece length': 16 * 1024,
                'pieces': 'aaaaaaaaaaaaaaaaaaaa'}}, {'max_buffer_size': 1}))

    def test_info_section(self):
        ti = lt.torrent_info('base.torrent')

        self.assertTrue(len(ti.info_section()) != 0)
        self.assertTrue(len(ti.hash_for_piece(0)) != 0)

    def test_torrent_info_bytes_overload(self):
        # bytes will never be interpreted as a file name. It's interpreted as a
        # bencoded buffer
        with self.assertRaises(RuntimeError):
            ti = lt.torrent_info(b'base.torrent')

    def test_web_seeds(self):
        ti = lt.torrent_info('base.torrent')

        ws = [{'url': 'http://foo/test', 'auth': '', 'type': 0},
              {'url': 'http://bar/test', 'auth': '', 'type': 1}]
        ti.set_web_seeds(ws)
        web_seeds = ti.web_seeds()
        self.assertEqual(len(ws), len(web_seeds))
        for i in range(len(web_seeds)):
            self.assertEqual(web_seeds[i]["url"], ws[i]["url"])
            self.assertEqual(web_seeds[i]["auth"], ws[i]["auth"])
            self.assertEqual(web_seeds[i]["type"], ws[i]["type"])

    def test_announce_entry(self):
        ae = lt.announce_entry('test')
        self.assertEqual(ae.url, 'test')
        self.assertEqual(ae.tier, 0)
        self.assertEqual(ae.verified, False)
        self.assertEqual(ae.source, 0)

    def test_torrent_info_sha1_overload(self):
        ti = lt.torrent_info(lt.info_hash_t(lt.sha1_hash(b'a' * 20)))
        self.assertEqual(ti.info_hash(), lt.sha1_hash(b'a' * 20))
        self.assertEqual(ti.info_hashes().v1, lt.sha1_hash(b'a' * 20))

        ti_copy = lt.torrent_info(ti)
        self.assertEqual(ti_copy.info_hash(), lt.sha1_hash(b'a' * 20))
        self.assertEqual(ti_copy.info_hashes().v1, lt.sha1_hash(b'a' * 20))

    def test_torrent_info_sha256_overload(self):
        ti = lt.torrent_info(lt.info_hash_t(lt.sha256_hash(b'a' * 32)))
        self.assertEqual(ti.info_hashes().v2, lt.sha256_hash(b'a' * 32))

        ti_copy = lt.torrent_info(ti)
        self.assertEqual(ti_copy.info_hashes().v2, lt.sha256_hash(b'a' * 32))

    def test_url_seed(self):
        ti = lt.torrent_info('base.torrent')

        ti.add_tracker('foobar1')
        ti.add_url_seed('foobar2')
        ti.add_url_seed('foobar3', 'username:password')
        ti.add_url_seed('foobar4', 'username:password', [])

        seeds = ti.web_seeds()
        self.assertEqual(seeds, [
            {'url': 'foobar2', 'type': 0, 'auth': ''},
            {'url': 'foobar3', 'type': 0, 'auth': 'username:password'},
            {'url': 'foobar4', 'type': 0, 'auth': 'username:password'},
        ])

    def test_http_seed(self):
        ti = lt.torrent_info('base.torrent')

        ti.add_http_seed('foobar2')
        ti.add_http_seed('foobar3', 'username:password')
        ti.add_http_seed('foobar4', 'username:password', [])

        seeds = ti.web_seeds()
        self.assertEqual(seeds, [
            {'url': 'foobar2', 'type': 1, 'auth': ''},
            {'url': 'foobar3', 'type': 1, 'auth': 'username:password'},
            {'url': 'foobar4', 'type': 1, 'auth': 'username:password'},
        ])

class test_alerts(unittest.TestCase):

    def test_alert(self):

        ses = lt.session(settings)
        ti = lt.torrent_info('base.torrent')
        h = ses.add_torrent({'ti': ti, 'save_path': os.getcwd()})
        st = h.status()
        time.sleep(1)
        ses.remove_torrent(h)
        ses.wait_for_alert(1000)  # milliseconds
        alerts = ses.pop_alerts()
        for a in alerts:
            if a.what() == 'add_torrent_alert':
                self.assertEqual(a.torrent_name, 'temp')
            print(a.message())
            for field_name in dir(a):
                if field_name.startswith('__'):
                    continue
                field = getattr(a, field_name)
                if callable(field):
                    print('  ', field_name, ' = ', field())
                else:
                    print('  ', field_name, ' = ', field)

        print(st.next_announce)
        self.assertEqual(st.name, 'temp')
        print(st.errc.message())
        print(st.pieces)
        print(st.last_seen_complete)
        print(st.completed_time)
        print(st.progress)
        print(st.num_pieces)
        print(st.distributed_copies)
        print(st.info_hashes)
        print(st.seeding_duration)
        print(st.last_upload)
        print(st.last_download)
        self.assertEqual(st.save_path, os.getcwd())

    def test_alert_fs(self):
        ses = lt.session(settings)
        s1, s2 = socket.socketpair()
        ses.set_alert_fd(s2.fileno())

        ses.pop_alerts()

        # make sure there's an alert to wake us up
        ses.post_session_stats()

        read_sockets, write_sockets, error_sockets = select.select([s1], [], [])

        self.assertEqual(len(read_sockets), 1)
        for s in read_sockets:
            s.recv(10)

    def test_pop_alerts(self):
        ses = lt.session(settings)
        ses.async_add_torrent(
            {"ti": lt.torrent_info("base.torrent"), "save_path": "."})

# this will cause an error (because of duplicate torrents) and the
# torrent_info object created here will be deleted once the alert goes out
# of scope. When that happens, it will decrement the python object, to allow
# it to release the object.
# we're trying to catch the error described in this post, with regards to
# torrent_info.
# https://mail.python.org/pipermail/cplusplus-sig/2007-June/012130.html
        ses.async_add_torrent(
            {"ti": lt.torrent_info("base.torrent"), "save_path": "."})
        time.sleep(1)
        for i in range(0, 10):
            alerts = ses.pop_alerts()
            for a in alerts:
                print(a.message())
            time.sleep(0.1)

    def test_alert_notify(self):
        ses = lt.session(settings)
        event = threading.Event()

        def callback():
            event.set()

        ses.set_alert_notify(callback)
        ses.async_add_torrent(
            {"ti": lt.torrent_info("base.torrent"), "save_path": "."})
        event.wait()


class test_bencoder(unittest.TestCase):

    def test_bencode(self):
        encoded = lt.bencode({'a': 1, 'b': [1, 2, 3], 'c': 'foo'})
        self.assertEqual(encoded, b'd1:ai1e1:bli1ei2ei3ee1:c3:fooe')

    def test_bdecode(self):
        encoded = b'd1:ai1e1:bli1ei2ei3ee1:c3:fooe'
        decoded = lt.bdecode(encoded)
        self.assertEqual(decoded, {b'a': 1, b'b': [1, 2, 3], b'c': b'foo'})

    def test_string(self):
        encoded = lt.bencode('foo\u00e5\u00e4\u00f6')
        self.assertEqual(encoded, b'9:foo\xc3\xa5\xc3\xa4\xc3\xb6')

    def test_bytes(self):
        encoded = lt.bencode(b'foo')
        self.assertEqual(encoded, b'3:foo')

    def test_float(self):
        # TODO: this should throw a TypeError in the future
        with self.assertWarns(DeprecationWarning):
            encoded = lt.bencode(1.337)
            self.assertEqual(encoded, b'0:')

    def test_object(self):
        class FooBar:
            dummy = 1

        # TODO: this should throw a TypeError in the future
        with self.assertWarns(DeprecationWarning):
            encoded = lt.bencode(FooBar())
            self.assertEqual(encoded, b'0:')

    def test_preformatted(self):
        encoded = lt.bencode((1, 2, 3, 4, 5))
        self.assertEqual(encoded, b'\x01\x02\x03\x04\x05')

class test_sha1hash(unittest.TestCase):

    def test_sha1hash(self):
        h = 'a0' * 20
        s = lt.sha1_hash(binascii.unhexlify(h))
        self.assertEqual(h, str(s))

    def test_hash(self):
        self.assertNotEqual(hash(lt.sha1_hash(b'b' * 20)), hash(lt.sha1_hash(b'a' * 20)))
        self.assertEqual(hash(lt.sha1_hash(b'b' * 20)), hash(lt.sha1_hash(b'b' * 20)))

class test_sha256hash(unittest.TestCase):

    def test_sha1hash(self):
        h = 'a0' * 32
        s = lt.sha256_hash(binascii.unhexlify(h))
        self.assertEqual(h, str(s))

    def test_hash(self):
        self.assertNotEqual(hash(lt.sha256_hash(b'b' * 32)), hash(lt.sha256_hash(b'a' * 32)))
        self.assertEqual(hash(lt.sha256_hash(b'b' * 32)), hash(lt.sha256_hash(b'b' * 32)))

class test_info_hash(unittest.TestCase):

    def test_info_hash(self):
        s1 = lt.sha1_hash(b'a' * 20)
        s2 = lt.sha256_hash(b'b' * 32)

        ih1 = lt.info_hash_t(s1);
        self.assertTrue(ih1.has_v1())
        self.assertFalse(ih1.has_v2())
        self.assertEqual(ih1.v1, s1)

        ih2 = lt.info_hash_t(s2);
        self.assertFalse(ih2.has_v1())
        self.assertTrue(ih2.has_v2())
        self.assertEqual(ih2.v2, s2)

        ih12 = lt.info_hash_t(s1, s2);
        self.assertTrue(ih12.has_v1())
        self.assertTrue(ih12.has_v2())
        self.assertEqual(ih12.v1, s1)
        self.assertEqual(ih12.v2, s2)

        self.assertNotEqual(hash(ih1), hash(ih2))
        self.assertNotEqual(hash(ih1), hash(ih12))
        self.assertEqual(hash(ih1), hash(lt.info_hash_t(s1)))
        self.assertEqual(hash(ih2), hash(lt.info_hash_t(s2)))
        self.assertEqual(hash(ih12), hash(lt.info_hash_t(s1, s2)))

class test_magnet_link(unittest.TestCase):

    def test_parse_magnet_uri(self):
        ses = lt.session({})
        magnet = 'magnet:?xt=urn:btih:C6EIF4CCYDBTIJVG3APAGM7M4NDONCTI'
        p = lt.parse_magnet_uri(magnet)
        self.assertEqual(str(p.info_hashes.v1), '178882f042c0c33426a6d81e0333ece346e68a68')
        p.save_path = '.'
        h = ses.add_torrent(p)
        self.assertEqual(str(h.info_hash()), '178882f042c0c33426a6d81e0333ece346e68a68')
        self.assertEqual(str(h.info_hashes().v1), '178882f042c0c33426a6d81e0333ece346e68a68')

    def test_parse_magnet_uri_dict(self):
        ses = lt.session({})
        magnet = 'magnet:?xt=urn:btih:C6EIF4CCYDBTIJVG3APAGM7M4NDONCTI'
        p = lt.parse_magnet_uri_dict(magnet)
        self.assertEqual(binascii.hexlify(p['info_hashes']), b'178882f042c0c33426a6d81e0333ece346e68a68')
        p['save_path'] = '.'
        h = ses.add_torrent(p)
        self.assertEqual(str(h.info_hash()), '178882f042c0c33426a6d81e0333ece346e68a68')
        self.assertEqual(str(h.info_hashes().v1), '178882f042c0c33426a6d81e0333ece346e68a68')

    def test_add_deprecated_magnet_link(self):
        ses = lt.session()
        atp = lt.add_torrent_params()
        atp.info_hashes = lt.info_hash_t(lt.sha1_hash(b"a" * 20))
        atp.save_path = "."
        h = ses.add_torrent(atp)

        self.assertTrue(h.status().info_hashes == lt.info_hash_t(lt.sha1_hash(b"a" * 20)))

    def test_add_magnet_link(self):
        ses = lt.session()
        atp = lt.add_torrent_params()
        atp.save_path = "."
        atp.info_hash = lt.sha1_hash(b"a" * 20)
        h = ses.add_torrent(atp)

        self.assertTrue(h.status().info_hashes == lt.info_hash_t(lt.sha1_hash(b"a" * 20)))


class test_peer_class(unittest.TestCase):

    def test_peer_class_ids(self):
        s = lt.session(settings)

        print('global_peer_class_id:', lt.session.global_peer_class_id)
        print('tcp_peer_class_id:', lt.session.tcp_peer_class_id)
        print('local_peer_class_id:', lt.session.local_peer_class_id)

        print('global: ', s.get_peer_class(s.global_peer_class_id))
        print('tcp: ', s.get_peer_class(s.local_peer_class_id))
        print('local: ', s.get_peer_class(s.local_peer_class_id))

    def test_peer_class(self):
        s = lt.session(settings)

        c = s.create_peer_class('test class')
        print('new class: ', s.get_peer_class(c))

        nfo = s.get_peer_class(c)
        self.assertEqual(nfo['download_limit'], 0)
        self.assertEqual(nfo['upload_limit'], 0)
        self.assertEqual(nfo['ignore_unchoke_slots'], False)
        self.assertEqual(nfo['connection_limit_factor'], 100)
        self.assertEqual(nfo['download_priority'], 1)
        self.assertEqual(nfo['upload_priority'], 1)
        self.assertEqual(nfo['label'], 'test class')

        nfo['download_limit'] = 1337
        nfo['upload_limit'] = 1338
        nfo['ignore_unchoke_slots'] = True
        nfo['connection_limit_factor'] = 42
        nfo['download_priority'] = 2
        nfo['upload_priority'] = 3

        s.set_peer_class(c, nfo)

        nfo2 = s.get_peer_class(c)
        self.assertEqual(nfo, nfo2)

    def test_peer_class_filter(self):
        filt = lt.peer_class_type_filter()
        filt.add(lt.peer_class_type_filter.tcp_socket, lt.session.global_peer_class_id)
        filt.remove(lt.peer_class_type_filter.utp_socket, lt.session.local_peer_class_id)

        filt.disallow(lt.peer_class_type_filter.tcp_socket, lt.session.global_peer_class_id)
        filt.allow(lt.peer_class_type_filter.utp_socket, lt.session.local_peer_class_id)

    def test_peer_class_ip_filter(self):
        s = lt.session(settings)
        s.set_peer_class_type_filter(lt.peer_class_type_filter())
        s.set_peer_class_filter(lt.ip_filter())

class test_ip_filter(unittest.TestCase):

    def test_export(self):

        f = lt.ip_filter()
        self.assertEqual(f.access('1.1.1.1'), 0)
        f.add_rule('1.1.1.1', '1.1.1.2', 1)
        self.assertEqual(f.access('1.1.1.0'), 0)
        self.assertEqual(f.access('1.1.1.1'), 1)
        self.assertEqual(f.access('1.1.1.2'), 1)
        self.assertEqual(f.access('1.1.1.3'), 0)
        exp = f.export_filter()
        self.assertEqual(exp, ([('0.0.0.0', '1.1.1.0'), ('1.1.1.1', '1.1.1.2'), ('1.1.1.3', '255.255.255.255')], [('::', 'ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff')]))

class test_session(unittest.TestCase):

    def test_settings(self):
        sett = { 'alert_mask': lt.alert.category_t.all_categories }
        s = lt.session(sett)
        sett = s.get_settings()
        self.assertEqual(sett['alert_mask'] & 0x7fffffff, 0x7fffffff)

    def test_session_params(self):
        sp = lt.session_params()
        sp.settings = { 'alert_mask': lt.alert.category_t.all_categories }
        s = lt.session(sp)
        sett = s.get_settings()
        self.assertEqual(sett['alert_mask'] & 0x7fffffff, 0x7fffffff)

    def test_session_params_constructor(self):
        sp = lt.session_params({ 'alert_mask': lt.alert.category_t.all_categories })
        s = lt.session(sp)
        sett = s.get_settings()
        self.assertEqual(sett['alert_mask'] & 0x7fffffff, 0x7fffffff)

    def test_session_params_ip_filter(self):
        sp = lt.session_params()
        sp.ip_filter.add_rule("1.1.1.1", "1.1.1.2", 1337)
        self.assertEqual(sp.ip_filter.access("1.1.1.1"), 1337)
        self.assertEqual(sp.ip_filter.access("1.1.1.2"), 1337)
        self.assertEqual(sp.ip_filter.access("1.1.1.3"), 0)

    def test_session_params_roundtrip_buf(self):

        sp = lt.session_params()
        sp.settings = { 'alert_mask': lt.alert.category_t.all_categories }

        buf = lt.write_session_params_buf(sp)
        sp2 = lt.read_session_params(buf)
        self.assertEqual(sp2.settings['alert_mask'] & 0x7fffffff, 0x7fffffff)

    def test_session_params_roundtrip_entry(self):

        sp = lt.session_params()
        sp.settings = { 'alert_mask': lt.alert.category_t.all_categories }

        ent = lt.write_session_params(sp)
        print(ent)
        sp2 = lt.read_session_params(ent)
        self.assertEqual(sp2.settings['alert_mask'] & 0x7fffffff, 0x7fffffff)

    def test_add_torrent(self):
        s = lt.session(settings)
        h = s.add_torrent({'ti': lt.torrent_info('base.torrent'),
                       'save_path': '.',
                       'dht_nodes': [('1.2.3.4', 6881), ('4.3.2.1', 6881)],
                       'http_seeds': ['http://test.com/seed'],
                       'peers': [('5.6.7.8', 6881)],
                       'banned_peers': [('8.7.6.5', 6881)],
                       'file_priorities': [1, 1, 1, 2, 0]})

    def test_find_torrent(self):
        s = lt.session(settings)
        h = s.add_torrent({'info_hash': b"a" * 20,
                           'save_path': '.'})
        self.assertTrue(h.is_valid())

        h2 = s.find_torrent(lt.sha1_hash(b"a" * 20))
        self.assertTrue(h2.is_valid())
        h3 = s.find_torrent(lt.sha1_hash(b"b" * 20))
        self.assertFalse(h3.is_valid())

        self.assertEqual(h, h2)
        self.assertNotEqual(h, h3)

    def test_add_torrent_info_hash(self):
        s = lt.session(settings)
        h = s.add_torrent({
                           'info_hash': b'a' * 20,
                           'info_hashes': b'a' * 32,
                           'save_path': '.'})

        time.sleep(1)
        alerts = s.pop_alerts()

        while len(alerts) > 0:
            a = alerts.pop(0)
            print(a)

        self.assertTrue(h.is_valid())
        self.assertEqual(h.status().info_hashes, lt.info_hash_t(lt.sha256_hash(b'a' * 32)))

    def test_session_status(self):
        if not has_deprecated():
            return

        s = lt.session()
        st = s.status()
        print(st)
        print(st.active_requests)
        print(st.dht_nodes)
        print(st.dht_node_cache)
        print(st.dht_torrents)
        print(st.dht_global_nodes)
        print(st.dht_total_allocations)

    def test_apply_settings(self):

        s = lt.session(settings)
        s.apply_settings({'num_want': 66, 'user_agent': 'test123'})
        self.assertEqual(s.get_settings()['num_want'], 66)
        self.assertEqual(s.get_settings()['user_agent'], 'test123')

    def test_post_session_stats(self):
        s = lt.session({'alert_mask': 0, 'enable_dht': False})
        s.post_session_stats()
        alerts = []
        # first the stats headers log line. but not if logging is disabled
        while len(alerts) == 0:
            s.wait_for_alert(1000)
            alerts = s.pop_alerts()

        while len(alerts) > 0:
            a = alerts.pop(0)
            print(a)
            if isinstance(a, lt.session_stats_header_alert):
                break
        self.assertTrue(isinstance(a, lt.session_stats_header_alert))
        # then the actual stats values
        while len(alerts) == 0:
            s.wait_for_alert(1000)
            alerts = s.pop_alerts()
        a = alerts.pop(0)
        print(a)
        self.assertTrue(isinstance(a, lt.session_stats_alert))
        self.assertTrue(isinstance(a.values, dict))
        self.assertTrue(len(a.values) > 0)

    def test_post_dht_stats(self):
        s = lt.session({'alert_mask': 0, 'enable_dht': False})
        s.post_dht_stats()
        alerts = []
        cnt = 0
        while len(alerts) == 0:
            s.wait_for_alert(1000)
            alerts = s.pop_alerts()
            cnt += 1
            if cnt > 60:
                print('no dht_stats_alert in 1 minute!')
                sys.exit(1)
        a = alerts.pop(0)
        self.assertTrue(isinstance(a, lt.dht_stats_alert))
        self.assertTrue(isinstance(a.active_requests, list))
        self.assertTrue(isinstance(a.routing_table, list))

    def test_unknown_settings(self):
        try:
            lt.session({'unexpected-key-name': 42})
            self.assertFalse('should have thrown an exception')
        except KeyError as e:
            print(e)

    def test_fingerprint(self):
        self.assertEqual(lt.generate_fingerprint('LT', 0, 1, 2, 3), '-LT0123-')
        self.assertEqual(lt.generate_fingerprint('..', 10, 1, 2, 3), '-..A123-')

    def test_min_memory_preset(self):
        min_mem = lt.min_memory_usage()
        print(min_mem)

        self.assertTrue('allow_idna' not in min_mem)
        self.assertTrue('connection_speed' in min_mem)
        self.assertTrue('file_pool_size' in min_mem)

    def test_seed_mode_preset(self):
        seed_mode = lt.high_performance_seed()
        print(seed_mode)

        self.assertTrue('alert_queue_size' in seed_mode)
        self.assertTrue('connection_speed' in seed_mode)
        self.assertTrue('file_pool_size' in seed_mode)

    def test_default_settings(self):

        default = lt.default_settings()
        print(default)


class test_example_client(unittest.TestCase):

    def test_execute_client(self):
        if os.name == 'nt':
            # TODO: fix windows includes of client.py
            return
        my_stdin = sys.stdin
        if os.name != 'nt':
            master_fd, slave_fd = pty.openpty()
            # slave_fd fix multiple stdin assignment at termios.tcgetattr
            my_stdin = slave_fd

        process = sub.Popen(
            [sys.executable, "client.py", "url_seed_multi.torrent"],
            stdin=my_stdin, stdout=sub.PIPE, stderr=sub.PIPE)
        # python2 has no Popen.wait() timeout
        time.sleep(5)
        returncode = process.poll()
        if returncode is None:
            # this is an expected use-case
            process.kill()
        err = process.stderr.read().decode("utf-8")
        self.assertEqual('', err, 'process throw errors: \n' + err)
        # check error code if process did unexpected end
        if returncode is not None:
            # in case of error return: output stdout if nothing was on stderr
            if returncode != 0:
                print("stdout:\n" + process.stdout.read().decode("utf-8"))
            self.assertEqual(returncode, 0, "returncode: " + str(returncode) + "\n"
                             + "stderr: empty\n"
                             + "some configuration does not output errors like missing module members,"
                             + "try to call it manually to get the error message\n")

    def test_execute_simple_client(self):
        process = sub.Popen(
            [sys.executable, "simple_client.py", "url_seed_multi.torrent"],
            stdout=sub.PIPE, stderr=sub.PIPE)
        # python2 has no Popen.wait() timeout
        time.sleep(5)
        returncode = process.poll()
        if returncode is None:
            # this is an expected use-case
            process.kill()
        err = process.stderr.read().decode("utf-8")
        self.assertEqual('', err, 'process throw errors: \n' + err)
        # check error code if process did unexpected end
        if returncode is not None:
            # in case of error return: output stdout if nothing was on stderr
            if returncode != 0:
                print("stdout:\n" + process.stdout.read().decode("utf-8"))
            self.assertEqual(returncode, 0, "returncode: " + str(returncode) + "\n"
                             + "stderr: empty\n"
                             + "some configuration does not output errors like missing module members,"
                             + "try to call it manually to get the error message\n")

    def test_execute_make_torrent(self):
        process = sub.Popen(
            [sys.executable, "make_torrent.py", "url_seed_multi.torrent",
             "http://test.com/test"], stdout=sub.PIPE, stderr=sub.PIPE)
        returncode = process.wait()
        # python2 has no Popen.wait() timeout
        err = process.stderr.read().decode("utf-8")
        self.assertEqual('', err, 'process throw errors: \n' + err)
        # in case of error return: output stdout if nothing was on stderr
        if returncode != 0:
            print("stdout:\n" + process.stdout.read().decode("utf-8"))
        self.assertEqual(returncode, 0, "returncode: " + str(returncode) + "\n"
                         + "stderr: empty\n"
                         + "some configuration does not output errors like missing module members,"
                         + "try to call it manually to get the error message\n")

    def test_default_settings(self):

        default = lt.default_settings()
        self.assertNotIn('', default)
        print(default)


class test_operation_t(unittest.TestCase):

    def test_enum(self):
        self.assertEqual(lt.operation_name(lt.operation_t.sock_accept), "sock_accept")
        self.assertEqual(lt.operation_name(lt.operation_t.unknown), "unknown")
        self.assertEqual(lt.operation_name(lt.operation_t.mkdir), "mkdir")
        self.assertEqual(lt.operation_name(lt.operation_t.partfile_write), "partfile_write")
        self.assertEqual(lt.operation_name(lt.operation_t.hostname_lookup), "hostname_lookup")


class test_error_code(unittest.TestCase):

    def test_error_code(self):

        a = lt.error_code()
        a = lt.error_code(10, lt.libtorrent_category())
        self.assertEqual(a.category().name(), 'libtorrent')

        self.assertEqual(lt.libtorrent_category().name(), 'libtorrent')
        self.assertEqual(lt.upnp_category().name(), 'upnp')
        self.assertEqual(lt.http_category().name(), 'http')
        self.assertEqual(lt.socks_category().name(), 'socks')
        self.assertEqual(lt.bdecode_category().name(), 'bdecode')
        self.assertEqual(lt.generic_category().name(), 'generic')
        self.assertEqual(lt.system_category().name(), 'system')


class test_peer_info(unittest.TestCase):

    def test_peer_info_members(self):

        p = lt.peer_info()

        print(p.client)
        print(p.pieces)
        print(p.pieces)
        print(p.last_request)
        print(p.last_active)
        print(p.flags)
        print(p.source)
        print(p.pid)
        print(p.downloading_piece_index)
        print(p.ip)
        print(p.local_endpoint)
        print(p.read_state)
        print(p.write_state)


class test_dht_settings(unittest.TestCase):

    def test_dht_get_peers(self):
        session = lt.session();
        info_hash = lt.sha1_hash(b"a" * 20)
        session.dht_get_peers(info_hash);

    def test_construct(self):

        ds = lt.dht_settings()

        print(ds.max_peers_reply)
        print(ds.search_branching)
        print(ds.max_fail_count)
        print(ds.max_fail_count)
        print(ds.max_torrents)
        print(ds.max_dht_items)
        print(ds.restrict_routing_ips)
        print(ds.restrict_search_ips)
        print(ds.max_torrent_search_reply)
        print(ds.extended_routing_table)
        print(ds.aggressive_lookups)
        print(ds.privacy_lookups)
        print(ds.enforce_node_id)
        print(ds.ignore_dark_internet)
        print(ds.block_timeout)
        print(ds.block_ratelimit)
        print(ds.read_only)
        print(ds.item_lifetime)


def get_isolated_settings():
    return {
        "enable_dht": False,
        "enable_lsd": False,
        "enable_natpmp": False,
        "enable_upnp": False,
        "listen_interfaces": "127.0.0.1:0",
        "dht_bootstrap_nodes": "",
    }


def loop_until_timeout(timeout, msg="condition"):
    deadline = time.monotonic() + timeout
    while time.monotonic() < deadline:
        yield
    raise AssertionError(f"{msg} timed out")


def unlink_all_files(path):
    for dirpath, _, filenames in os.walk(path):
        for filename in filenames:
            filepath = os.path.join(dirpath, filename)
            os.unlink(filepath)


# In test cases where libtorrent writes torrent data in a temporary directory,
# cleaning up the tempdir on Windows CI sometimes fails with a PermissionError
# having WinError 5 (Access Denied). I can't repro this WinError in any way;
# holding an open file handle results in a different WinError. Seems to be a
# race condition which only happens with very short-lived tests which write
# data. Work around by cleaning up the tempdir in a loop.

# TODO: why is this necessary?
def cleanup_with_windows_fix(tempdir, *, timeout):
    # Clean up just the files, so we don't have to bother with depth-first
    # traversal
    for _ in loop_until_timeout(timeout, msg="PermissionError clear"):
        try:
            unlink_all_files(tempdir.name)
        except PermissionError:
            if sys.platform == "win32":
                # current release of mypy doesn't know about winerror
                # if exc.winerror == 5:
                continue
            raise
        break
    # This removes directories in depth-first traversal.
    # It also marks the tempdir as explicitly cleaned so it doesn't trigger a
    # ResourceWarning.
    tempdir.cleanup()


def wait_for(session, alert_type, *, timeout, prefix=None):
    # Return the first alert of type, but log all alerts.
    result = None
    for _ in loop_until_timeout(timeout, msg=alert_type.__name__):
        for alert in session.pop_alerts():
            print(f"{alert.what()}: {alert.message()}")
            if result is None and isinstance(alert, alert_type):
                result = alert
        if result is not None:
            return result
    raise AssertionError("unreachable")


class LambdaRequestHandler(http.server.BaseHTTPRequestHandler):
    default_request_version = "HTTP/1.1"

    def __init__(self, get_data, *args, **kwargs):
        self.get_data = get_data
        super().__init__(*args, **kwargs)

    def do_GET(self):
        print(f"mock tracker request: {self.requestline}")
        data = self.get_data()
        self.send_response(200)
        self.send_header("Content-Type", "application/octet-stream")
        self.send_header("Content-Length", str(len(data)))
        self.end_headers()
        self.wfile.write(data)


class SSLTrackerAlertTest(unittest.TestCase):

    def setUp(self):
        self.cert_path = os.path.realpath(os.path.join(
            os.path.dirname(__file__), "..", "..", "test", "ssl", "server.pem"
        ))
        print(f"cert_path = {self.cert_path}")

        self.tracker_response = {
            b"external ip": b"\x01\x02\x03\x04",
        }
        self.tracker = http.server.HTTPServer(
            ("127.0.0.1", 0),
            functools.partial(
                LambdaRequestHandler, lambda: lt.bencode(self.tracker_response)
            ),
        )
        self.ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.ctx.load_cert_chain(self.cert_path)
        self.tracker.socket = self.ctx.wrap_socket(
            self.tracker.socket, server_side=True
        )
        self.tracker_thread = threading.Thread(target=self.tracker.serve_forever)
        self.tracker_thread.start()
        # HTTPServer.server_name seems to resolve to things like
        # "localhost.localdomain"
        port = self.tracker.server_port
        self.tracker_url = f"https://127.0.0.1:{port}/announce"
        print(f"mock tracker url = {self.tracker_url}")

        self.settings = get_isolated_settings()
        self.settings["alert_mask"] = lt.alert_category.status
        # I couldn't get validation to work on all platforms. Setting
        # SSL_CERT_FILE to our self-signed cert works on linux and mac, but
        # not on Windows.
        self.settings["validate_https_trackers"] = False
        self.session = lt.session(self.settings)
        self.dir = tempfile.TemporaryDirectory()
        self.atp = lt.add_torrent_params()
        self.atp.info_hash = dummy_data.get_sha1_hash()
        self.atp.flags &= ~lt.torrent_flags.auto_managed
        self.atp.flags &= ~lt.torrent_flags.paused
        self.atp.save_path = self.dir.name

    def tearDown(self):
        # we do this because sessions writing data can collide with
        # cleaning up temporary directories. session.abort() isn't bound
        handles = self.session.get_torrents()
        for handle in handles:
            self.session.remove_torrent(handle)
        for _ in loop_until_timeout(5, msg="clear all handles"):
            if not any(handle.is_valid() for handle in handles):
                break
        cleanup_with_windows_fix(self.dir, timeout=5)
        self.tracker.shutdown()
        # Explicitly clean up server sockets, to avoid ResourceWarning
        self.tracker.server_close()

    def test_external_ip_alert_via_ssl_tracker(self):
        handle = self.session.add_torrent(self.atp)
        handle.add_tracker({"url": self.tracker_url})

        alert = wait_for(self.session, lt.external_ip_alert, timeout=60)

        self.assertEqual(alert.category(), lt.alert_category.status)
        self.assertEqual(alert.what(), "external_ip")
        self.assertIsInstance(alert.message(), str)
        self.assertNotEqual(alert.message(), "")
        self.assertEqual(str(alert), alert.message())
        self.assertEqual(alert.external_address, "1.2.3.4")


if __name__ == '__main__':
    print(lt.__version__)
    try:
        shutil.copy(os.path.join('..', '..', 'test', 'test_torrents',
                                 'url_seed_multi.torrent'), '.')
    except shutil.SameFileError:
        pass
    try:
        shutil.copy(os.path.join('..', '..', 'test', 'test_torrents',
                                 'base.torrent'), '.')
    except shutil.SameFileError:
        pass
    try:
        shutil.copy(os.path.join('..', '..', 'test', 'test_torrents',
                                 'unordered.torrent'), '.')
    except shutil.SameFileError:
        pass
    unittest.main()



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

# this is meant to parse the dht_lookups.log generated by parse_dht_log.py
from __future__ import print_function

import os

nodes = {}


def get_origin(n):
    if n in nodes:
        return list(nodes[n]['conns'])
    else:
        return ['0.0.0.0']


def calculate_pos(nid, dist):
    nid = int(nid[0:7], 16)

    x = 0
    y = 0
    for i in range(0, 28, 2):
        x |= (nid & (1 << i)) >> (i / 2)
        y |= (nid & (2 << i)) >> (i / 2 + 1)

# print '%d -> %d %d' % (dist, x, y)

    return (x / 3, y / 3)


def plot_nodes(nodes, frame):

    try:
        os.mkdir('dht_frames')
    except Exception:
        pass

    out = open('dht_frames/plot-%02d.dot' % frame, 'w+')
    edges = set()
    print('graph swarm {', file=out)
# print >>out, '"tl" [shape=point pos="0,0!"];'
# print >>out, '"tr" [shape=point pos="1638,0!"];'
# print >>out, '"ll" [shape=point pos="1638,1638!"];'
# print >>out, '"tr" [shape=point pos="0,1638!"];'
    for dst, n in list(nodes.items()):
        shape = 'point'
        if 's' in n:
            shape = n['s']

        print('"%s" [shape=%s fillcolor="%s" label="" pos="%d,%d!"];' %
              (dst, shape, n['c'], n['p'][0], n['p'][1]), file=out)
        for e in n['conns']:
            if (e, dst) in edges:
                continue

            # only add an edge once to the .dot file
            edges.add((e, dst))
            edges.add((dst, e))

            style = 'solid'
            col = 'gray'
            if nodes[dst]['c'] != 'white' and nodes[e]['c'] != 'white':
                style = 'solid'
                col = 'black'
            print('"%s" -- "%s" [style="%s" color="%s"];' % (e, dst, style, col), file=out)

    print('}', file=out)
    out.close()
    os.system('neato -n dht_frames/plot-%02d.dot -Tpng -o dht_frames/frame-%02d.png' % (frame, frame))


frame = 0
next_render_time = 100
f = open('dht_lookups.txt')
for line in f:
    if line.startswith('***'):
        break

    kind = line[0:3].strip()
    line = line[3:].strip().split(' ')

    if kind == '===':
        continue

    t = int(line[0])
    if t > next_render_time:
        plot_nodes(nodes, frame)
        frame += 1
        next_render_time += 100
        # sys.exit(0)

    if kind == '<>':
        p = calculate_pos(line[1], 0)
        dst = '0.0.0.0'
        if dst not in nodes:
            nodes[dst] = {'conns': set(), 'p': p, 'c': 'blue', 's': 'circle'}

        p = calculate_pos(line[2], 25)
        dst = '255.255.255.255'
        if dst not in nodes:
            nodes[dst] = {'conns': set(), 'p': p, 'c': 'yellow', 's': 'circle'}
    elif kind == '->':
        dst = line[3]

        if dst not in nodes:
            src = get_origin(dst)
            p = calculate_pos(line[2], int(line[1]))
            nodes[dst] = {'conns': set(src), 'p': p, 'c': 'grey'}
        nodes[dst]['c'] = 'grey'

    elif kind == '+':
        dst = line[3]
        src = line[4]
        p = calculate_pos(line[2], int(line[1]))
        if dst not in nodes:
            nodes[dst] = {'conns': set(), 'p': p, 'c': 'white'}
        nodes[dst]['conns'].add(src)

    elif kind == '<-':
        dst = line[3]
        nodes[dst]['c'] = 'green'
    elif kind == 'x':
        dst = line[3]
        nodes[dst]['c'] = 'orange'
    elif kind == 'X':
        dst = line[3]
        nodes[dst]['c'] = 'red'

f.close()



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

from __future__ import print_function

# Copyright (c) 2016, Arvid Norberg
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#     * Neither the name of the author nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# this script can parse and generate reports from the alert log from a
# libtorrent session

import os
import sys
import math
from multiprocessing.pool import ThreadPool

thread_pool = ThreadPool(8)

output_dir = 'session_stats_report'

stat = open(sys.argv[1])
line = stat.readline()
print('looking for stats header')
while 'session stats header:' not in line:
    line = stat.readline()

print('found')
keys = line.split('session stats header:')[1].strip().split(', ')

try:
    os.mkdir(output_dir)
except Exception:
    pass
data_out = open(os.path.join(output_dir, 'counters.dat'), 'w+')

idx = 0
for line in stat:
    if 'session stats (' not in line:
        continue
    data_out.write(("%d\t" % idx) + line.split(' values): ')[1].strip().replace(', ', '\t') + '\n')
    idx += 1

data_out.close()

line_graph = 0
histogram = 1
stacked = 2
diff = 3

graph_colors = []

pattern = [[0, 0, 1], [0, 1, 0], [1, 0, 0], [1, 0, 1], [0, 1, 1], [1, 1, 0]]


def process_color(c, op):
    for i in range(3):
        if op == 0:
            c[i] = min(255, c[i] + 0xb0)
        if op == 2:
            c[i] = max(0, c[i] - 0x50)
    return c


for i in range(0, len(pattern) * 3):

    op = i / len(pattern)

    c = list(pattern[i % len(pattern)])
    for j in range(3):
        c[j] *= 0xff
    c = process_color(c, op)

    c = '#%02x%02x%02x' % (c[0], c[1], c[2])
    graph_colors.append(c)

line_colors = list(graph_colors)
line_colors.reverse()

gradient16_colors = []
for i in range(0, 16):
    f = i / 16.
    pi = 3.1415927
    r = max(int(255 * (math.sin(f * pi) + 0.2)), 0)
    g = max(int(255 * (math.sin((f - 0.5) * pi) + 0.2)), 0)
    b = max(int(255 * (math.sin((f + 0.5) * pi) + 0.2)), 0)
    c = '#%02x%02x%02x' % (min(r, 255), min(g, 255), min(b, 255))
    gradient16_colors.append(c)

gradient18_colors = []
for i in range(0, 18):
    f = i / 18.
    pi = 3.1415927
    r = max(int(255 * (math.sin(f * pi) + 0.2)), 0)
    g = max(int(255 * (math.sin((f - 0.5) * pi) + 0.2)), 0)
    b = max(int(255 * (math.sin((f + 0.5) * pi) + 0.2)), 0)
    c = '#%02x%02x%02x' % (min(r, 255), min(g, 255), min(b, 255))
    gradient18_colors.append(c)

gradient6_colors = []
for i in range(0, 6):
    f = i / 6.
    c = '#%02x%02x%02x' % (min(int(255 * (-2 * f + 2)), 255), min(int(255 * (2 * f)), 255), 100)
    gradient6_colors.append(c)


def plot_fun(script):
    try:
        ret = os.system('gnuplot "%s" 2>/dev/null' % script)
    except Exception as e:
        print('please install gnuplot: sudo apt install gnuplot')
        raise e
    if ret != 0 and ret != 256:
        print('gnuplot failed: %d\n' % ret)
        raise Exception("abort")

    sys.stdout.write('.')
    sys.stdout.flush()


def to_title(key):
    return key.replace('_', ' ').replace('.', ' - ')


def gen_report(name, unit, lines, short_unit, generation, log_file, options):
    filename = os.path.join(output_dir, '%s_%04d.png' % (name, generation))
    thumb = os.path.join(output_dir, '%s_%04d_thumb.png' % (name, generation))

    # don't re-render a graph unless the logfile has changed
    try:
        dst1 = os.stat(filename)
        dst2 = os.stat(thumb)
        src = os.stat(log_file)

        if dst1.st_mtime > src.st_mtime and dst2.st_mtime > src.st_mtime:
            sys.stdout.write('.')
            return None

    except Exception:
        pass

    script = os.path.join(output_dir, '%s_%04d.gnuplot' % (name, generation))
    out = open(script, 'w')
    print("set term png size 1200,700", file=out)
    print('set output "%s"' % filename, file=out)
    if 'allow-negative' not in options:
        print('set yrange [0:*]', file=out)
    print("set tics nomirror", file=out)
    print("set key box", file=out)
    print("set key left top", file=out)

    colors = graph_colors
    if options['type'] == line_graph:
        colors = line_colors

    try:
        if options['colors'] == 'gradient16':
            colors = gradient16_colors
        elif options['colors'] == 'gradient6':
            colors = gradient6_colors
        if options['colors'] == 'gradient18':
            colors = gradient18_colors
    except Exception:
        pass

    if options['type'] == histogram:
        binwidth = options['binwidth']
        numbins = int(options['numbins'])

        print('binwidth=%f' % binwidth, file=out)
        print('set boxwidth binwidth', file=out)
        print('bin(x,width)=width*floor(x/width) + binwidth/2', file=out)
        print('set xrange [0:%f]' % (binwidth * numbins), file=out)
        print('set xlabel "%s"' % unit, file=out)
        print('set ylabel "number"', file=out)

        k = lines[0]
        try:
            column = keys.index(k) + 2
        except Exception:
            print('"%s" not found' % k)
            return
        print('plot "%s" using (bin($%d,binwidth)):(1.0) smooth freq with boxes' % (log_file, column), file=out)
        print('', file=out)
        print('', file=out)
        print('', file=out)

    elif options['type'] == stacked:
        print('set xrange [0:*]', file=out)
        print('set ylabel "%s"' % unit, file=out)
        print('set xlabel "time (s)"', file=out)
        print('set format y "%%.1s%%c%s";' % short_unit, file=out)
        print('set style fill solid 1.0 noborder', file=out)
        print('plot', end=' ', file=out)
        first = True
        graph = ''
        plot_expression = ''
        color = 0
        for k in lines:
            try:
                column = keys.index(k) + 2
            except Exception:
                print('"%s" not found' % k)
                continue
            if not first:
                plot_expression = ', ' + plot_expression
                graph += '+'
            axis = 'x1y1'
            graph += '$%d' % column
            plot_expression = ' "%s" using 1:(%s) title "%s" axes %s with filledcurves x1 lc rgb "%s"' % (
                log_file, graph, to_title(k), axis, colors[color % len(colors)]) + plot_expression
            first = False
            color += 1
        print(plot_expression, file=out)
    elif options['type'] == diff:
        print('set xrange [0:*]', file=out)
        print('set ylabel "%s"' % unit, file=out)
        print('set xlabel "time (s)"', file=out)
        print('set format y "%%.1s%%c%s";' % short_unit, file=out)
        first = True
        graph = ''
        title = ''
        for k in lines:
            try:
                column = keys.index(k) + 2
            except Exception:
                print('"%s" not found' % k)
                continue
            if not first:
                graph += '-'
                title += ' - '
            graph += '$%d' % column
            title += to_title(k)
            first = False
        print('plot "%s" using 1:(%s) title "%s" with step' % (log_file, graph, title), file=out)
    else:
        print('set xrange [0:*]', file=out)
        print('set ylabel "%s"' % unit, file=out)
        print('set xlabel "time (s)"', file=out)
        print('set format y "%%.1s%%c%s";' % short_unit, file=out)
        print('plot', end=' ', file=out)
        first = True
        color = 0
        for k in lines:
            try:
                column = keys.index(k) + 2
            except Exception:
                print('"%s" not found' % k)
                continue
            if not first:
                print(', ', end=' ', file=out)
            axis = 'x1y1'
            print(' "%s" using 1:%d title "%s" axes %s with steps lc rgb "%s"' %
                  (log_file, column, to_title(k), axis, colors[color % len(colors)]), end=' ', file=out)
            first = False
            color += 1
        print('', file=out)

    print('set term png size 150,100', file=out)
    print('set output "%s"' % thumb, file=out)
    print('set key off', file=out)
    print('unset tics', file=out)
    print('set format x ""', file=out)
    print('set format y ""', file=out)
    print('set xlabel ""', file=out)
    print('set ylabel ""', file=out)
    print('set y2label ""', file=out)
    print('set rmargin 0', file=out)
    print('set lmargin 0', file=out)
    print('set tmargin 0', file=out)
    print('set bmargin 0', file=out)
    print("replot", file=out)
    out.close()
    return script


def gen_html(reports, generations):
    file = open(os.path.join(output_dir, 'index.html'), 'w+')

    css = '''img { margin: 0}
#head { display: block }
#graphs { white-space:nowrap; }
h1 { line-height: 1; display: inline }
h2 { line-height: 1; display: inline; font-size: 1em; font-weight: normal};'''

    print('<html><head><style type="text/css">%s</style></head><body>' % css, file=file)

    for i in reports:
        print('<div id="head"><h1>%s </h1><h2>%s</h2><div><div id="graphs">' % (i[0], i[3]), file=file)
        for g in generations:
            print('<a href="%s_%04d.png"><img src="%s_%04d_thumb.png"></a>' % (i[0], g, i[0], g), file=file)
        print('</div>', file=file)

    print('</body></html>', file=file)
    file.close()


reports = [

    ('torrents', 'num', '', 'number of torrents in different torrent states', [
        'ses.num_downloading_torrents',
        'ses.num_seeding_torrents',
        'ses.num_checking_torrents',
        'ses.num_stopped_torrents',
        'ses.num_upload_only_torrents',
        'ses.num_error_torrents',
        'ses.num_queued_seeding_torrents',
        'ses.num_queued_download_torrents'
    ], {'type': stacked}),

    ('peers', 'num', '', 'num connected peers', [
     'peer.num_peers_connected', 'peer.num_peers_half_open'], {'type': stacked}),
    ('peers_max', 'num', '', 'num connected peers', ['peer.num_peers_connected', 'peer.num_peers_half_open']),
    ('peer_churn', 'num', '', 'connecting and disconnecting peers',
     ['peer.num_peers_half_open', 'peer.connection_attempts', 'peer.boost_connection_attempts',
      'peer.missed_connection_attempts', 'peer.no_peer_connection_attempts']),
    ('new_peers', 'num', '', '', ['peer.incoming_connections', 'peer.connection_attempts']),
    ('connection_attempts', 'num', '', '', ['peer.connection_attempt_loops', 'peer.connection_attempts']),
    ('pieces', 'num', '', 'number completed pieces', [
     'ses.num_total_pieces_added', 'ses.num_piece_passed', 'ses.num_piece_failed']),
    ('disk_write_queue', 'Bytes', 'B', 'bytes queued up by peers, to be written to disk', ['disk.queued_write_bytes']),

    ('peers_requests', 'num', '', 'incoming piece request rate', [
        'peer.piece_requests',
        'peer.max_piece_requests',
        'peer.invalid_piece_requests',
        'peer.choked_piece_requests',
        'peer.cancelled_piece_requests'
    ]),

    ('peers_upload', 'num', '', 'number of peers by state wrt. uploading', [
        'peer.num_peers_up_disk',
        'peer.num_peers_up_interested',
        'peer.num_peers_up_unchoked_all',
        'peer.num_peers_up_unchoked_optimistic',
        'peer.num_peers_up_unchoked',
        'peer.num_peers_up_requests'
    ]),

    ('peers_download', 'num', '', 'number of peers by state wrt. downloading', [
        'peer.num_peers_down_interested',
        'peer.num_peers_down_unchoked',
        'peer.num_peers_down_requests',
        'peer.num_peers_down_disk'
    ]),

    ('peer_errors', 'num', '', 'number of peers by error that disconnected them', [
        'peer.disconnected_peers',
        'peer.eof_peers',
        'peer.connreset_peers',
        'peer.connrefused_peers',
        'peer.connaborted_peers',
        'peer.perm_peers',
        'peer.buffer_peers',
        'peer.unreachable_peers',
        'peer.broken_pipe_peers',
        'peer.addrinuse_peers',
        'peer.no_access_peers',
        'peer.invalid_arg_peers',
        'peer.aborted_peers'
    ], {'type': stacked}),

    ('peer_errors_incoming', 'num', '', 'number of peers by incoming or outgoing connection', [
        'peer.error_incoming_peers',
        'peer.error_outgoing_peers'
    ]),

    ('peer_errors_transport', 'num', '', 'number of peers by transport protocol', [
        'peer.error_tcp_peers',
        'peer.error_utp_peers'
    ]),

    ('peer_errors_encryption', 'num', '', 'number of peers by encryption level', [
        'peer.error_encrypted_peers',
        'peer.error_rc4_peers',
    ]),

    ('incoming requests', 'num', '', 'incoming 16kiB block requests', ['ses.num_incoming_request']),

    ('waste', 'downloaded bytes', 'B', 'proportion of all downloaded bytes that were wasted', [
        'net.recv_failed_bytes',
        'net.recv_redundant_bytes',
        'net.recv_ip_overhead_bytes'
    ], {'type': stacked}),

    ('waste by source', 'num wasted bytes', 'B', 'what is causing the waste', [
        'ses.waste_piece_timed_out',
        'ses.waste_piece_cancelled',
        'ses.waste_piece_unknown',
        'ses.waste_piece_seed',
        'ses.waste_piece_end_game',
        'ses.waste_piece_closing'
    ], {'type': stacked}),

    ('disk_time', '% of total disk job time', '%%', 'proportion of time spent by the disk thread',
     ['disk.disk_read_time', 'disk.disk_write_time', 'disk.disk_hash_time'], {'type': stacked}),
    ('disk_queue', 'blocks (16kiB)', '', 'disk store-buffer size',
     ['disk.num_write_jobs',
      'disk.num_read_jobs',
      'disk.num_jobs',
      'disk.queued_disk_jobs',
      'disk.blocked_disk_jobs']),
    ('disk fences', 'num', '', 'number of jobs currently blocked by a fence job', ['disk.blocked_disk_jobs']),
    ('fence jobs', 'num', '', 'active fence jobs per type', [
        'disk.num_fenced_move_storage',
        'disk.num_fenced_release_files',
        'disk.num_fenced_delete_files',
        'disk.num_fenced_check_fastresume',
        'disk.num_fenced_save_resume_data',
        'disk.num_fenced_rename_file',
        'disk.num_fenced_stop_torrent',
        'disk.num_fenced_file_priority',
        'disk.num_fenced_clear_piece'],
     {'type':stacked}),
    ('disk threads', 'num', '', 'number of disk threads currently writing',
     ['disk.num_writing_threads', 'disk.num_running_threads']),
    # ('mixed mode', 'rate', 'B/s', 'rates by transport protocol',
    #  ['TCP up rate','TCP down rate','uTP up rate','uTP down rate','TCP up limit','TCP down limit']),

    ('connection_type', 'num', '', 'peers by transport protocol', [ \
        'peer.num_tcp_peers', \
        'peer.num_socks5_peers', \
        'peer.num_http_proxy_peers', \
        'peer.num_utp_peers', \
        'peer.num_i2p_peers', \
        'peer.num_ssl_peers', \
        'peer.num_ssl_socks5_peers', \
        'peer.num_ssl_http_proxy_peers', \
        'peer.num_ssl_utp_peers' \
    ]),

    # ('uTP delay', 'buffering delay', 's', 'network delays measured by uTP',
    #  ['uTP peak send delay','uTP peak recv delay', 'uTP avg send delay', 'uTP avg recv delay']),
    # ('uTP send delay histogram', 'buffering delay', 's', 'send delays measured by uTP',
    #  ['uTP avg send delay'], {'type': histogram, 'binwidth': 0.05, 'numbins': 100}),
    # ('uTP recv delay histogram', 'buffering delay', 's', 'receive delays measured by uTP',
    #  ['uTP avg recv delay'], {'type': histogram, 'binwidth': 0.05, 'numbins': 100}),

    ('uTP stats', 'num', '', 'number of uTP events', [ \
        'utp.utp_packet_loss', \
        'utp.utp_timeout', \
        'utp.utp_packets_in', \
        'utp.utp_packets_out', \
        'utp.utp_fast_retransmit', \
        'utp.utp_packet_resend', \
        'utp.utp_samples_above_target', \
        'utp.utp_samples_below_target', \
        'utp.utp_payload_pkts_in', \
        'utp.utp_payload_pkts_out', \
        'utp.utp_invalid_pkts_in', \
        'utp.utp_redundant_pkts_in' \
    ], {'type': stacked}),

    ('boost.asio messages', 'num events', '', 'number of messages posted', [ \
        'net.on_read_counter', \
        'net.on_write_counter', \
        'net.on_tick_counter', \
        'net.on_lsd_counter', \
        'net.on_lsd_peer_counter', \
        'net.on_udp_counter', \
        'net.on_accept_counter', \
        'net.on_disk_counter' \
    ], {'type': stacked}),

    ('send_buffer_sizes', 'num', '', '', [ \
        'sock_bufs.socket_send_size3', \
        'sock_bufs.socket_send_size4', \
        'sock_bufs.socket_send_size5', \
        'sock_bufs.socket_send_size6', \
        'sock_bufs.socket_send_size7', \
        'sock_bufs.socket_send_size8', \
        'sock_bufs.socket_send_size9', \
        'sock_bufs.socket_send_size10', \
        'sock_bufs.socket_send_size11', \
        'sock_bufs.socket_send_size12', \
        'sock_bufs.socket_send_size13', \
        'sock_bufs.socket_send_size14', \
        'sock_bufs.socket_send_size15', \
        'sock_bufs.socket_send_size16', \
        'sock_bufs.socket_send_size17', \
        'sock_bufs.socket_send_size18', \
        'sock_bufs.socket_send_size19', \
        'sock_bufs.socket_send_size20' \
    ], {'type': stacked, 'colors': 'gradient18'}),

    ('recv_buffer_sizes', 'num', '', '', [ \
        'sock_bufs.socket_recv_size3', \
        'sock_bufs.socket_recv_size4', \
        'sock_bufs.socket_recv_size5', \
        'sock_bufs.socket_recv_size6', \
        'sock_bufs.socket_recv_size7', \
        'sock_bufs.socket_recv_size8', \
        'sock_bufs.socket_recv_size9', \
        'sock_bufs.socket_recv_size10', \
        'sock_bufs.socket_recv_size11', \
        'sock_bufs.socket_recv_size12', \
        'sock_bufs.socket_recv_size13', \
        'sock_bufs.socket_recv_size14', \
        'sock_bufs.socket_recv_size15', \
        'sock_bufs.socket_recv_size16', \
        'sock_bufs.socket_recv_size17', \
        'sock_bufs.socket_recv_size18', \
        'sock_bufs.socket_recv_size19', \
        'sock_bufs.socket_recv_size20' \
    ], {'type': stacked, 'colors': 'gradient18'}),

    ('request latency', 'us', '', 'latency from receiving requests to sending response', ['disk.request_latency']),
    ('incoming messages', 'num', '', 'number of received bittorrent messages, by type', [ \
        'ses.num_incoming_choke', \
        'ses.num_incoming_unchoke', \
        'ses.num_incoming_interested', \
        'ses.num_incoming_not_interested', \
        'ses.num_incoming_have', \
        'ses.num_incoming_bitfield', \
        'ses.num_incoming_request', \
        'ses.num_incoming_piece', \
        'ses.num_incoming_cancel', \
        'ses.num_incoming_dht_port', \
        'ses.num_incoming_suggest', \
        'ses.num_incoming_have_all', \
        'ses.num_incoming_have_none', \
        'ses.num_incoming_reject', \
        'ses.num_incoming_allowed_fast', \
        'ses.num_incoming_ext_handshake', \
        'ses.num_incoming_pex', \
        'ses.num_incoming_metadata', \
        'ses.num_incoming_extended' \
    ], {'type': stacked}),
    ('outgoing messages', 'num', '', 'number of sent bittorrent messages, by type', [ \
        'ses.num_outgoing_choke', \
        'ses.num_outgoing_unchoke', \
        'ses.num_outgoing_interested', \
        'ses.num_outgoing_not_interested', \
        'ses.num_outgoing_have', \
        'ses.num_outgoing_bitfield', \
        'ses.num_outgoing_request', \
        'ses.num_outgoing_piece', \
        'ses.num_outgoing_cancel', \
        'ses.num_outgoing_dht_port', \
        'ses.num_outgoing_suggest', \
        'ses.num_outgoing_have_all', \
        'ses.num_outgoing_have_none', \
        'ses.num_outgoing_reject', \
        'ses.num_outgoing_allowed_fast', \
        'ses.num_outgoing_ext_handshake', \
        'ses.num_outgoing_pex', \
        'ses.num_outgoing_metadata', \
        'ses.num_outgoing_extended' \
    ], {'type': stacked}),
    ('request in balance', 'num', '', 'request and piece message balance', [ \
        'ses.num_incoming_request', \
        'ses.num_outgoing_piece', \
        'ses.num_outgoing_reject', \
    ], {'type': diff}),
    ('request out balance', 'num', '', 'request and piece message balance', [ \
        'ses.num_outgoing_request', \
        'ses.num_incoming_piece', \
        'ses.num_incoming_reject', \
    ], {'type': diff}),

    # somewhat uninteresting stats
    # ('peer_dl_rates', 'num', '', 'peers split into download rate buckets',
    #  ['peers down 0', 'peers down 0-2', 'peers down 2-5', 'peers down 5-10', 'peers down 50-100', 'peers down 100-'],
    #  {'type':stacked, 'colors':'gradient6'}),
    # ('peer_dl_rates2', 'num', '', 'peers split into download rate buckets (only downloading peers)',
    #  ['peers down 0-2', 'peers down 2-5', 'peers down 5-10', 'peers down 50-100', 'peers down 100-'],
    #  {'type':stacked, 'colors':'gradient6'}),
    # ('peer_ul_rates', 'num', '', 'peers split into upload rate buckets',
    #  ['peers up 0', 'peers up 0-2', 'peers up 2-5', 'peers up 5-10', 'peers up 50-100', 'peers up 100-'],
    #  {'type':stacked, 'colors':'gradient6'}),
    # ('peer_ul_rates2', 'num', '', 'peers split into upload rate buckets (only uploading peers)',
    #  ['peers up 0-2', 'peers up 2-5', 'peers up 5-10', 'peers up 50-100', 'peers up 100-'],
    #  {'type':stacked, 'colors':'gradient6'}),

    ('piece_picker_invocations', 'invocations of piece picker', '', '', [ \
        'picker.reject_piece_picks', \
        'picker.unchoke_piece_picks', \
        'picker.incoming_redundant_piece_picks', \
        'picker.incoming_piece_picks', \
        'picker.end_game_piece_picks', \
        'picker.snubbed_piece_picks', \
        'picker.interesting_piece_picks', \
        'picker.hash_fail_piece_picks' \
    ], {'type': stacked}),
    ('piece_picker_loops', 'loops through piece picker', '', '', [ \
        'picker.piece_picker_partial_loops', \
        'picker.piece_picker_suggest_loops', \
        'picker.piece_picker_sequential_loops', \
        'picker.piece_picker_reverse_rare_loops', \
        'picker.piece_picker_rare_loops', \
        'picker.piece_picker_rand_start_loops', \
        'picker.piece_picker_rand_loops', \
        'picker.piece_picker_busy_loops' \
    ], {'type': stacked}),
    ('async_accept', 'number of outstanding accept calls', '', '', [ \
        'ses.num_outstanding_accept' \
    ]),
    ('queued_trackers', 'number of queued tracker announces', '', '', [ \
        'tracker.num_queued_tracker_announces' \
    ]),

    # ('picker_full_partials_distribution', 'full pieces', '', '', ['num full partial pieces'],
    #  {'type': histogram, 'binwidth': 5, 'numbins': 120}),
    # ('picker_partials_distribution', 'partial pieces', '', '', ['num downloading partial pieces'],
    #  {'type': histogram, 'binwidth': 5, 'numbins': 120})
]

print('generating graphs')
g = 0
generations = []
scripts = []

print('[%s] %04d\r[' % (' ' * len(reports), g), end='')
for i in reports:
    try:
        options = i[5]
    except Exception:
        options = {}
    if 'type' not in options:
        options['type'] = line_graph

    script = gen_report(i[0], i[1], i[4], i[2], g, os.path.join(output_dir, 'counters.dat'), options)
    if script is not None:
        scripts.append(script)

generations.append(g)
g += 1

# run gnuplot on all scripts, in parallel
thread_pool.map(plot_fun, scripts)
scripts = []

print('\ngenerating html')
gen_html(reports, generations)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import sys

# to use this script, first run 'sample' to sample your libtorrent based process
# the output can then be passed to this script to auto-fold call stacks at
# relevant depths and to filter out low sample counts
f = open(sys.argv[1])


def parse_line(line):
    indentation = 0
    while indentation < len(line) and line[indentation] == ' ':
        indentation += 1
    if indentation == 0:
        return (0, 0, '')

    line = line.strip().split(' ')
    samples = int(line[0])
    fun = ' '.join(line[1:])

    return (indentation, samples, fun)


fold = -1

try:
    sample_limit = int(sys.argv[2])
except Exception:
    sample_limit = 5

fun_samples = {}

for line in f:
    if 'Sort by top of stack' in line:
        break

    indentation, samples, fun = parse_line(line)
    if samples < sample_limit:
        continue
    if fold != -1 and indentation > fold:
        continue
    fold = -1

    if '__gnu_cxx::__normal_iterator<' in fun:
        fold = indentation - 1
        continue

    if 'boost::_bi::bind_t' in fun:
        continue
    if 'boost::_bi::list' in fun:
        continue
    if 'boost::_mfi::mf' in fun:
        continue
    if 'boost::_bi::storage' in fun:
        continue

# should only add leaves
    if fun in fun_samples:
        fun_samples[fun] += samples
    else:
        fun_samples[fun] = samples

    output = '%s%-4d %s' % (' ' * (indentation / 2), samples, fun)
    if len(output) > 200:
        output = output[0:200]
    print(output)

    if 'invariant_checker_impl' in fun:
        fold = indentation
    if 'free_multiple_buffers' in fun:
        fold = indentation
    if 'libtorrent::condition::wait' in fun:
        fold = indentation
    if 'allocate_buffer' in fun:
        fold = indentation
    if '::find_POD' in fun:
        fold = indentation
    if 'SHA1_Update' in fun:
        fold = indentation
    if 'boost::detail::function::basic_vtable' in fun:
        fold = indentation
    if 'operator new' in fun:
        fold = indentation
    if 'malloc' == fun:
        fold = indentation
    if 'free' == fun:
        fold = indentation
    if 'std::_Rb_tree' in fun:
        fold = indentation
    if 'pthread_cond_wait' in fun:
        fold = indentation
    if 'mp_exptmod' == fun:
        fold = indentation
    if '::check_invariant()' in fun:
        fold = indentation
    if 'libtorrent::condition::wait' in fun:
        fold = indentation
    if '_sigtramp' in fun:
        fold = indentation
    if 'time_now_hires' in fun:
        fold = indentation
    if 'libtorrent::sleep' in fun:
        fold = indentation
    if 'puts' == fun:
        fold = indentation
    if 'boost::asio::basic_stream_socket' in fun:
        fold = indentation
    if 'recvmsg' == fun:
        fold = indentation
    if 'sendmsg' == fun:
        fold = indentation
    if 'semaphore_signal_trap' == fun:
        fold = indentation
    if 'boost::detail::atomic_count::operator' in fun:
        fold = indentation
    if 'pthread_mutex_lock' == fun:
        fold = indentation
    if 'pthread_mutex_unlock' == fun:
        fold = indentation
    if '>::~vector()' == fun:
        fold = indentation
    if 'szone_free_definite_size' == fun:
        fold = indentation
    if 'snprintf' == fun:
        fold = indentation
    if 'usleep' == fun:
        fold = indentation
    if 'pthread_mutex_lock' == fun:
        fold = indentation
    if 'pthread_mutex_unlock' == fun:
        fold = indentation
    if 'std::string::append' in fun:
        fold = indentation
    if 'getipnodebyname' == fun:
        fold = indentation
    if '__gnu_debug::_Safe_iterator<std::' in fun:
        fold = indentation
    if 'fflush' == fun:
        fold = indentation
    if 'vfprintf' == fun:
        fold = indentation
    if 'fprintf' == fun:
        fold = indentation
    if 'BN_mod_exp' == fun:
        fold = indentation
    if 'BN_CTX_free' == fun:
        fold = indentation
    if 'cerror' == fun:
        fold = indentation
    if '0xffffffff' == fun:
        fold = indentation

list = []
for k in fun_samples:
    list.append((fun_samples[k], k))

list = sorted(list, reverse=True)

for i in list:
    print('%-4d %s' % (i[0], i[1]))



#!/usr/bin/env python3


import sys
import os
import libtorrent

if len(sys.argv) < 3:
    print('usage make_torrent.py file tracker-url')
    sys.exit(1)

input = os.path.abspath(sys.argv[1])

fs = libtorrent.file_storage()

# def predicate(f):
#   print f
#   return True
# libtorrent.add_files(fs, input, predicate)

parent_input = os.path.split(input)[0]

# if we have a single file, use it because os.walk does not work on a single files
if os.path.isfile(input):
    size = os.path.getsize(input)
    fs.add_file(input, size)

for root, dirs, files in os.walk(input):
    # skip directories starting with .
    if os.path.split(root)[1][0] == '.':
        continue

    for f in files:
        # skip files starting with .
        if f[0] == '.':
            continue

        # skip thumbs.db on windows
        if f == 'Thumbs.db':
            continue

        fname = os.path.join(root[len(parent_input) + 1:], f)
        size = os.path.getsize(os.path.join(parent_input, fname))
        print('%10d kiB  %s' % (size / 1024, fname))
        fs.add_file(fname, size)

if fs.num_files() == 0:
    print('no files added')
    sys.exit(1)

t = libtorrent.create_torrent(fs, 0, 4 * 1024 * 1024)

t.add_tracker(sys.argv[2])
t.set_creator('libtorrent %s' % libtorrent.__version__)

libtorrent.set_piece_hashes(t, parent_input, lambda x: sys.stdout.write('.'))
sys.stdout.write('\n')

f = open('out.torrent', 'wb+')
f.write(libtorrent.bencode(t.generate()))
f.close()



import sys
import os
import string
import hashlib

if len(sys.argv) < 2:
    print('usage: unify_corpus_names.py <corpus-directory>\n')
    sys.exit(1)

root = sys.argv[1]
for name in os.listdir(root):
    f = os.path.join(root, name)

    # ignore directories
    if not os.path.isfile(f):
        continue

    # if the name already looks like a SHA-1 hash, ignore it
    if len(name) == 40 and all(c in string.hexdigits for c in name):
        continue

    new_name = hashlib.sha1(open(f, 'rb').read()).hexdigest()
    print('%s -> %s' % (f, new_name))
    os.rename(f, os.path.join(root, new_name))



#!/usr/bin/env python3

# Copyright Daniel Wallin 2006. Use, modification and distribution is
# subject to the Boost Software License, Version 1.0. (See accompanying
# file LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)


import sys
import atexit
import libtorrent as lt
import time
import os.path


class WindowsConsole:
    def __init__(self):
        self.console = Console.getconsole()

    def clear(self):
        self.console.page()

    def write(self, str):
        self.console.write(str)

    def sleep_and_input(self, seconds):
        time.sleep(seconds)
        if msvcrt.kbhit():
            return msvcrt.getch()
        return None


class UnixConsole:
    def __init__(self):
        self.fd = sys.stdin
        self.old = termios.tcgetattr(self.fd.fileno())
        new = termios.tcgetattr(self.fd.fileno())
        new[3] = new[3] & ~termios.ICANON
        new[6][termios.VTIME] = 0
        new[6][termios.VMIN] = 1
        termios.tcsetattr(self.fd.fileno(), termios.TCSADRAIN, new)

        atexit.register(self._onexit)

    def _onexit(self):
        termios.tcsetattr(self.fd.fileno(), termios.TCSADRAIN, self.old)

    def clear(self):
        sys.stdout.write('\033[2J\033[0;0H')
        sys.stdout.flush()

    def write(self, str):
        sys.stdout.write(str)
        sys.stdout.flush()

    def sleep_and_input(self, seconds):
        read, __, __ = select.select(
            [self.fd.fileno()], [], [], seconds)
        if len(read) > 0:
            return self.fd.read(1)
        return None


if os.name == 'nt':
    import Console
    import msvcrt
else:
    import termios
    import select


def write_line(console, line):
    console.write(line)


def add_suffix(val):
    prefix = ['B', 'kB', 'MB', 'GB', 'TB']
    for i in range(len(prefix)):
        if abs(val) < 1000:
            if i == 0:
                return '%5.3g%s' % (val, prefix[i])
            else:
                return '%4.3g%s' % (val, prefix[i])
        val /= 1000

    return '%6.3gPB' % val


def progress_bar(progress, width):
    assert(progress <= 1)
    progress_chars = int(progress * width + 0.5)
    return progress_chars * '#' + (width - progress_chars) * '-'


def print_peer_info(console, peers):

    out = (' down    (total )   up     (total )'
           '  q  r flags  block progress  client\n')

    for p in peers:

        out += '%s/s ' % add_suffix(p.down_speed)
        out += '(%s) ' % add_suffix(p.total_download)
        out += '%s/s ' % add_suffix(p.up_speed)
        out += '(%s) ' % add_suffix(p.total_upload)
        out += '%2d ' % p.download_queue_length
        out += '%2d ' % p.upload_queue_length

        out += 'I' if p.flags & lt.peer_info.interesting else '.'
        out += 'C' if p.flags & lt.peer_info.choked else '.'
        out += 'i' if p.flags & lt.peer_info.remote_interested else '.'
        out += 'c' if p.flags & lt.peer_info.remote_choked else '.'
        out += 'e' if p.flags & lt.peer_info.supports_extensions else '.'
        out += 'l' if p.flags & lt.peer_info.local_connection else 'r'
        out += ' '

        if p.downloading_piece_index >= 0:
            assert(p.downloading_progress <= p.downloading_total)
            out += progress_bar(float(p.downloading_progress) /
                                p.downloading_total, 15)
        else:
            out += progress_bar(0, 15)
        out += ' '

        if p.flags & lt.peer_info.handshake:
            id = 'waiting for handshake'
        elif p.flags & lt.peer_info.connecting:
            id = 'connecting to peer'
        else:
            id = p.client

        out += '%s\n' % id[:10]

    write_line(console, out)


def print_download_queue(console, download_queue):

    out = ""

    for e in download_queue:
        out += '%4d: [' % e['piece_index']
        for b in e['blocks']:
            s = b['state']
            if s == 3:
                out += '#'
            elif s == 2:
                out += '='
            elif s == 1:
                out += '-'
            else:
                out += ' '
        out += ']\n'

    write_line(console, out)


def add_torrent(ses, filename, options):
    atp = lt.add_torrent_params()
    if filename.startswith('magnet:'):
        atp = lt.parse_magnet_uri(filename)
    else:
        ti = lt.torrent_info(filename)
        resume_file = os.path.join(options.save_path, ti.name() + '.fastresume')
        try:
            atp = lt.read_resume_data(open(resume_file, 'rb').read())
        except Exception as e:
            print('failed to open resume file "%s": %s' % (resume_file, e))
        atp.ti = ti

    atp.save_path = options.save_path
    atp.storage_mode = lt.storage_mode_t.storage_mode_sparse
    atp.flags |= lt.torrent_flags.duplicate_is_error \
        | lt.torrent_flags.auto_managed \
        | lt.torrent_flags.duplicate_is_error
    ses.async_add_torrent(atp)


def main():
    from optparse import OptionParser

    parser = OptionParser()

    parser.add_option('-p', '--port', type='int', help='set listening port')

    parser.add_option(
        '-i', '--listen-interface', type='string',
        help='set interface for incoming connections', )

    parser.add_option(
        '-o', '--outgoing-interface', type='string',
        help='set interface for outgoing connections')

    parser.add_option(
        '-d', '--max-download-rate', type='float',
        help='the maximum download rate given in kB/s. 0 means infinite.')

    parser.add_option(
        '-u', '--max-upload-rate', type='float',
        help='the maximum upload rate given in kB/s. 0 means infinite.')

    parser.add_option(
        '-s', '--save-path', type='string',
        help='the path where the downloaded file/folder should be placed.')

    parser.add_option(
        '-r', '--proxy-host', type='string',
        help='sets HTTP proxy host and port (separated by \':\')')

    parser.set_defaults(
        port=6881,
        listen_interface='0.0.0.0',
        outgoing_interface='',
        max_download_rate=0,
        max_upload_rate=0,
        save_path='.',
        proxy_host=''
    )

    (options, args) = parser.parse_args()

    if options.port < 0 or options.port > 65525:
        options.port = 6881

    options.max_upload_rate *= 1000
    options.max_download_rate *= 1000

    if options.max_upload_rate <= 0:
        options.max_upload_rate = -1
    if options.max_download_rate <= 0:
        options.max_download_rate = -1

    settings = {
        'user_agent': 'python_client/' + lt.__version__,
        'listen_interfaces': '%s:%d' % (options.listen_interface, options.port),
        'download_rate_limit': int(options.max_download_rate),
        'upload_rate_limit': int(options.max_upload_rate),
        'alert_mask': lt.alert.category_t.all_categories,
        'outgoing_interfaces': options.outgoing_interface,
    }

    if options.proxy_host != '':
        settings['proxy_hostname'] = options.proxy_host.split(':')[0]
        settings['proxy_type'] = lt.proxy_type_t.http
        settings['proxy_port'] = options.proxy_host.split(':')[1]

    ses = lt.session(settings)

    # map torrent_handle to torrent_status
    torrents = {}
    alerts_log = []

    for f in args:
        add_torrent(ses, f, options)

    if os.name == 'nt':
        console = WindowsConsole()
    else:
        console = UnixConsole()

    alive = True
    while alive:
        console.clear()

        out = ''

        for h, t in torrents.items():
            out += 'name: %-40s\n' % t.name[:40]

            if t.state != lt.torrent_status.seeding:
                state_str = ['queued', 'checking', 'downloading metadata',
                             'downloading', 'finished', 'seeding',
                             '', 'checking fastresume']
                out += state_str[t.state] + ' '

                out += '%5.4f%% ' % (t.progress * 100)
                out += progress_bar(t.progress, 49)
                out += '\n'

                out += 'total downloaded: %d Bytes\n' % t.total_done
                out += 'peers: %d seeds: %d distributed copies: %d\n' % \
                    (t.num_peers, t.num_seeds, t.distributed_copies)
                out += '\n'

            out += 'download: %s/s (%s) ' \
                % (add_suffix(t.download_rate), add_suffix(t.total_download))
            out += 'upload: %s/s (%s) ' \
                % (add_suffix(t.upload_rate), add_suffix(t.total_upload))

            if t.state != lt.torrent_status.seeding:
                out += 'info-hash: %s\n' % t.info_hashes
                out += 'next announce: %s\n' % t.next_announce
                out += 'tracker: %s\n' % t.current_tracker

            write_line(console, out)

            print_peer_info(console, t.handle.get_peer_info())
            print_download_queue(console, t.handle.get_download_queue())

            if t.state != lt.torrent_status.seeding:
                try:
                    out = '\n'
                    fp = h.file_progress()
                    ti = t.torrent_file
                    for idx, p in enumerate(fp):
                        out += progress_bar(p / float(ti.files().file_size(idx)), 20)
                        out += ' ' + ti.files().file_path(idx) + '\n'
                    write_line(console, out)
                except Exception:
                    pass

        write_line(console, 76 * '-' + '\n')
        write_line(console, '(q)uit), (p)ause), (u)npause), (r)eannounce\n')
        write_line(console, 76 * '-' + '\n')

        alerts = ses.pop_alerts()
        for a in alerts:
            alerts_log.append(a.message())

            # add new torrents to our list of torrent_status
            if isinstance(a, lt.add_torrent_alert):
                h = a.handle
                h.set_max_connections(60)
                h.set_max_uploads(-1)
                torrents[h] = h.status()

            # update our torrent_status array for torrents that have
            # changed some of their state
            if isinstance(a, lt.state_update_alert):
                for s in a.status:
                    torrents[s.handle] = s

        if len(alerts_log) > 20:
            alerts_log = alerts_log[-20:]

        for a in alerts_log:
            write_line(console, a + '\n')

        c = console.sleep_and_input(0.5)

        ses.post_torrent_updates()
        if not c:
            continue

        if c == 'r':
            for h in torrents:
                h.force_reannounce()
        elif c == 'q':
            alive = False
        elif c == 'p':
            for h in torrents:
                h.pause()
        elif c == 'u':
            for h in torrents:
                h.resume()

    ses.pause()
    for h, t in torrents.items():
        if not h.is_valid() or not t.has_metadata:
            continue
        h.save_resume_data()

    while len(torrents) > 0:
        alerts = ses.pop_alerts()
        for a in alerts:
            if isinstance(a, lt.save_resume_data_alert):
                print(a)
                data = lt.write_resume_data_buf(a.params)
                h = a.handle
                if h in torrents:
                    open(os.path.join(options.save_path, torrents[h].name + '.fastresume'), 'wb').write(data)
                    del torrents[h]

            if isinstance(a, lt.save_resume_data_failed_alert):
                h = a.handle
                if h in torrents:
                    print('failed to save resume data for ', torrents[h].name)
                    del torrents[h]
        time.sleep(0.5)


main()



import os
import shutil
import hashlib
import struct
import random

corpus_dirs = [
    'torrent_info', 'upnp', 'gzip', 'base32decode', 'base32encode',
    'base64encode', 'bdecode_node', 'convert_from_native', 'convert_to_native',
    'dht_node', 'escape_path', 'escape_string', 'file_storage_add_file',
    'http_parser', 'lazy_bdecode', 'parse_int', 'parse_magnet_uri', 'resume_data',
    'sanitize_path', 'utf8_codepoint', 'utp',
    'verify_encoding', 'peer_conn', 'add_torrent', 'idna', 'parse_url', 'http_tracker']

for p in corpus_dirs:
    try:
        os.makedirs(os.path.join('corpus', p))
    except Exception as e:
        print(e)

torrent_dir = '../test/test_torrents'
for f in os.listdir(torrent_dir):
    shutil.copy(os.path.join(torrent_dir, f), os.path.join('corpus', 'torrent_info'))

xml_tests = [
    '<a blah="b"></a>', '<a b=c></a>', '<a b"c"></a>', '<a b="c></a>',
    '<![CDATA[<sender>John Smith</sender>]]>', '<![CDATA[<sender>John S',
    '<!-- comment -->', '<empty></empty>', '<tag',
    '''<?xml version="1.0" encoding="ISO-8859-1" ?>
<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema"></xs:schema>''',
    '<selfclosing />']

for x in xml_tests:
    name = hashlib.sha1(x.encode('ascii')).hexdigest()
    with open(os.path.join('corpus', 'upnp', name), 'w+') as f:
        f.write(x)

gzip_dir = '../test'
for f in ['zeroes.gz', 'corrupt.gz', 'invalid1.gz']:
    shutil.copy(os.path.join(gzip_dir, f), os.path.join('corpus', 'gzip'))

idna = ['....', 'xn--foo-.bar', 'foo.xn--bar-.com', 'Xn--foobar-', 'XN--foobar-', '..xnxn--foobar-']

counter = 0
for i in idna:
    open(os.path.join('corpus', 'idna', '%d' % counter), 'w+').write(i)
    counter += 1

urls = ['https://user:password@example.com:8080/path?query']

counter = 0
for i in urls:
    open(os.path.join('corpus', 'parse_url', '%d' % counter), 'w+').write(i)
    counter += 1

counter = 0
tracker_fields = ['interval', 'min interval', 'tracker id', 'failure reason',
    'warning message', 'complete', 'incomplete', 'downloaded', 'downloaders', 'external ip']
tracker_values = ['i-1e', 'i0e', 'i1800e', '6:foobar', 'de', '0:', 'le']
peer_fields = ['peer id', 'ip', 'port']
peer_values = ['i-1e', 'i0e', 'i1800e', '6:foobar', 'de', '0:', 'le', '9:127.0.0.1']

for i in range(1000):
    tracker_msg = 'd'
    for f in tracker_fields:
        tracker_msg += '%d:' % len(f) + f
        tracker_msg += random.choice(tracker_values)

    tracker_msg += '5:filesd20:ababababababababababd'
    for f in tracker_fields:
        tracker_msg += '%d:' % len(f) + f
        tracker_msg += random.choice(tracker_values)
    tracker_msg += 'ee'

    tracker_msg += '5:peers'
    if random.getrandbits(1) == 0:
        tracker_msg += 'l'
        for k in range(10):
            tracker_msg += 'd'
            for f in peer_fields:
                tracker_msg += '%d:' % len(f) + f
                tracker_msg += random.choice(peer_values)
            tracker_msg += 'e'
        tracker_msg += 'e'
    else:
        tracker_msg += '60:'
        for k in range(6*10):
            tracker_msg += chr(random.getrandbits(8))

    tracker_msg += '6:peers6'
    tracker_msg += '180:'
    for k in range(18*10):
        tracker_msg += chr(random.getrandbits(8))

    tracker_msg += 'e'
    open(os.path.join('corpus', 'http_tracker', '%d' % counter), 'w+').write(tracker_msg)
    counter += 1

# generate peer protocol messages
messages = []


def add_length(msg):
    return struct.pack('>I', len(msg)) + msg


def add_reserved(msg):
    return b'\0\0\0\0\0\x18\0\x05' + msg


# extended handshake
def add_extended_handshake(msg):
    ext_handshake = b'd1:md11:ut_metadatai1e11:lt_donthavei2e12:ut_holepunch' + \
        b'i3e11:upload_onlyi4ee11:upload_onlyi1e10:share_modei1e4:reqqi1234e6:yourip4:0000e'
    return add_length(struct.pack('BB', 20, 0) + ext_handshake) + msg


# request
for i in range(101):
    for j in range(-1, 1):
        messages.append(add_length(struct.pack('>Biii', 6, i, j, 0x4000)))

# cancel
for i in range(101):
    for j in range(-1, 1):
        messages.append(add_length(struct.pack('>Biii', 8, i, j, 0x4000)))

# piece
for i in range(101):
    messages.append(add_length(struct.pack('>Bii', 7, i, 0) + (b'a' * 0x4000)))

# single-byte
for i in range(256):
    messages.append(add_length(struct.pack('B', i)))

# reject
for i in range(101):
    messages.append(add_length(struct.pack('>Biii', 16, i, 0, 0x4000)))

# suggest
for i in range(101):
    messages.append(add_length(struct.pack('>Bi', 13, i)))

# allow-fast
for i in range(101):
    messages.append(add_length(struct.pack('>Bi', 17, i)))

# have
for i in range(101):
    messages.append(add_length(struct.pack('>Bi', 4, i)))

# DHT-port
for i in range(101):
    messages.append(add_length(struct.pack('>BH', 9, i * 10)))

# hash request
for i in range(-10, 200, 20):
    for j in range(-1, 1):
        for k in range(-1, 1):
            for m in range(-1, 1):
                for n in range(-1, 1):
                    messages.append(add_length(struct.pack('>Biiiii', 21, i, j, k, m, n)))

# hash reject
for i in range(-10, 200, 20):
    for j in range(-1, 1):
        for k in range(-1, 1):
            for m in range(-1, 1):
                for n in range(-1, 1):
                    messages.append(add_length(struct.pack('>Biiiii', 23, i, j, k, m, n)))

# hash
for i in range(-10, 200, 20):
    for j in range(-1, 1):
        messages.append(add_length(struct.pack('>Biiiii', 22, i, j, 0, 2, 0) + (b'0' * 32 * 5)))

# lt_dont_have
messages.append(add_extended_handshake(add_length(struct.pack('>BBi', 20, 7, -1))))
messages.append(add_extended_handshake(add_length(struct.pack('>BBi', 20, 7, 0))))
messages.append(add_extended_handshake(add_length(struct.pack('>BBi', 20, 7, 0x7fffffff))))

# share mode
messages.append(add_extended_handshake(add_length(struct.pack('BBB', 20, 8, 255))))
messages.append(add_extended_handshake(add_length(struct.pack('BBB', 20, 8, 0))))
messages.append(add_extended_handshake(add_length(struct.pack('BBB', 20, 8, 1))))

# holepunch
for i in range(0, 2):
    for j in range(0, 1):
        messages.append(add_extended_handshake(add_length(struct.pack('>BBBBiH', 20, 4, i, j, 0, 0))))
        messages.append(add_extended_handshake(add_length(struct.pack('>BBBBiiH', 20, 4, i, j, 0, 0, 0))))

# upload only
for i in range(0, 1):
    messages.append(add_extended_handshake(add_length(struct.pack('BBB', 20, 3, i))))

# bitfields
bitfield_len = (100 + 7) // 8

for i in range(256):
    messages.append(add_length(struct.pack('B', 5) + (struct.pack('B', i) * bitfield_len)))

mixes = []

for i in range(200):
    random.shuffle(messages)
    mixes.append(b''.join(messages[1:20]))

messages += mixes

for m in messages:
    f = open('corpus/peer_conn/%s' % hashlib.sha1(m).hexdigest(), 'wb+')
    f.write(add_reserved(m))
    f.close()



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

from __future__ import print_function

import sys
import os
import pprint

pp = pprint.PrettyPrinter(indent=4)

up_time_quanta = 500

f = open(sys.argv[1])

announce_histogram = {}

# TODO: make this histogram into a CDF

node_uptime_histogram = {}

counter = 0

# maps search_id to a list of events. Each event is a dict containing:
#  t: timestamp
#  d: distance (from target)
#  o: outstanding searches
#  e: event (NEW, COMPLETED, ADD, INVOKE, TIMEOUT)
#  i: node-id
#  a: IP address and port
#  s: source node-id (only for ADD events)
outstanding_searches = {}

# list of completed searches
searches = []


def convert_timestamp(t):
    parts = t.split('.')
    hms = parts[0].split(':')
    return (int(hms[0]) * 3600 + int(hms[1]) * 60 + int(hms[2])) * 1000 + int(parts[1])


last_incoming = ''

our_node_id = ''

unique_ips = set()
client_version_histogram = {}
client_histogram = {}

for line in f:
    counter += 1
    # if counter % 1000 == 0:
    #     print '\r%d' % counter,
    try:
        ls = line.split(' ')
        if 'starting DHT tracker with node id:' in line:
            our_node_id = ls[ls.index('id:') + 1].strip()

        try:
            if len(ls) > 4 and ls[2] == '<==' and ls[1] == '[dht_tracker]':
                ip = ls[3].split(':')[0]
                if ip not in unique_ips:
                    unique_ips.add(ip)
                    json_blob = line.split(ls[3])[1]
                    version = json_blob.split("'v': '")[1].split("'")[0]
                    if len(version) == 4:
                        v = '%s-%d' % (version[0:2], (ord(version[2]) << 8) + ord(version[3]))
                    elif len(version) == 8:
                        v = '%c%c-%d' % (chr(int(version[0:2], 16)), chr(int(version[2:4], 16)), int(version[4:8], 16))
                    else:
                        v = 'unknown'

                    if v not in client_version_histogram:
                        client_version_histogram[v] = 1
                    else:
                        client_version_histogram[v] += 1

                    if not v[0:2] in client_histogram:
                        client_histogram[v[0:2]] = 1
                    else:
                        client_histogram[v[0:2]] += 1
        except Exception:
            pass

        if 'announce-distance:' in line:
            idx = ls.index('announce-distance:')

            d = int(ls[idx + 1].strip())
            if d not in announce_histogram:
                announce_histogram[d] = 0
            announce_histogram[d] += 1
        if 'NODE FAILED' in line:
            idx = ls.index('fails:')
            if int(ls[idx + 1].strip()) != 1:
                continue
            idx = ls.index('up-time:')
            d = int(ls[idx + 1].strip())
            # quantize
            d = d - (d % up_time_quanta)
            if d not in node_uptime_histogram:
                node_uptime_histogram[d] = 0
            node_uptime_histogram[d] += 1

        search_id = ls[2]
        ts = ls[0]
        event = ls[3]

        if event == 'RESPONSE':
            outstanding = int(ls[ls.index('invoke-count:') + 1])
            distance = int(ls[ls.index('distance:') + 1])
            nid = ls[ls.index('id:') + 1]
            addr = ls[ls.index('addr:') + 1]
            last_response = addr
            outstanding_searches[search_id].append({'t': ts, 'd': distance,
                                                    'o': outstanding + 1, 'a': addr, 'e': event, 'i': nid})
        elif event == 'NEW':
            nid = ls[ls.index('target:') + 1]
            outstanding_searches[search_id] = [{'t': ts, 'd': 0, 'o': 0,
                                                'e': event, 'abstime': ts, 'i': nid}]
            last_response = ''
        elif event == 'INVOKE' or event == 'ADD' or event == '1ST_TIMEOUT' or \
                event == 'TIMEOUT' or event == 'PEERS':
            if search_id not in outstanding_searches:
                print('orphaned event: %s' % line)
            else:
                outstanding = int(ls[ls.index('invoke-count:') + 1])
                distance = int(ls[ls.index('distance:') + 1])
                nid = ls[ls.index('id:') + 1]
                addr = ls[ls.index('addr:') + 1]
                source = ''
                if event == 'ADD':
                    if last_response == '':
                        continue
                    source = last_response

                outstanding_searches[search_id].append(
                    {'t': ts, 'd': distance, 'o': outstanding + 1, 'a': addr, 'e': event, 'i': nid, 's': source})
        elif event == 'ABORTED':
            outstanding_searches[search_id].append({'t': ts, 'e': event})
        elif event == 'COMPLETED':
            distance = int(ls[ls.index('distance:') + 1])
            lookup_type = ls[ls.index('type:') + 1].strip()
            outstanding_searches[search_id].append({'t': ts, 'd': distance,
                                                    'o': 0, 'e': event, 'i': ''})

            outstanding_searches[search_id][0]['type'] = lookup_type

            s = outstanding_searches[search_id]

            try:
                start_time = convert_timestamp(s[0]['t'])
                for i in range(len(s)):
                    s[i]['t'] = convert_timestamp(s[i]['t']) - start_time
            except Exception:
                pass
            searches.append(s)
            del outstanding_searches[search_id]

    except Exception as e:
        print(e)
        print(line.split(' '))

lookup_times_min = []
lookup_times_max = []

# these are the timestamps for lookups crossing distance
# to target boundaries
lookup_distance = []
for i in range(0, 15):
    lookup_distance.append([])

for s in searches:
    for i in s:
        if 'last_dist' not in i:
            i['last_dist'] = -1
        cur_dist = 160 - i['d']
        last_dist = i['last_dist']
        if cur_dist > last_dist:
            for j in range(last_dist + 1, cur_dist + 1):
                if j >= len(lookup_distance):
                    break
                lookup_distance[j].append(i['t'])
            i['last_dist'] = cur_dist
        if i['e'] != 'PEERS':
            continue
        lookup_times_min.append(i['t'])
        break
    for i in reversed(s):
        if i['e'] != 'PEERS':
            continue
        lookup_times_max.append(i['t'])
        break


lookup_times_min.sort()
lookup_times_max.sort()
out = open('dht_lookup_times_cdf.txt', 'w+')
counter = 0
for i in range(len(lookup_times_min)):
    counter += 1
    print('%d\t%d\t%f' % (lookup_times_min[i], lookup_times_max[i], counter / float(len(lookup_times_min))), file=out)
out.close()

for i in lookup_distance:
    i.sort()

dist = 0
for i in lookup_distance:
    out = open('dht_lookup_distance_%d.txt' % dist, 'w+')
    dist += 1
    counter = 0
    for j in i:
        counter += 1
        print('%d\t%f' % (j, counter / float(len(i))), file=out)
    out.close()

out = open('dht_lookups.txt', 'w+')
for s in searches:
    for i in s:
        if i['e'] == 'INVOKE':
            print(' ->', i['t'], 160 - i['d'], i['i'], i['a'], file=out)
        elif i['e'] == '1ST_TIMEOUT':
            print(' x ', i['t'], 160 - i['d'], i['i'], i['a'], file=out)
        elif i['e'] == 'TIMEOUT':
            print(' X ', i['t'], 160 - i['d'], i['i'], i['a'], file=out)
        elif i['e'] == 'ADD':
            print(' + ', i['t'], 160 - i['d'], i['i'], i['a'], i['s'], file=out)
        elif i['e'] == 'RESPONSE':
            print(' <-', i['t'], 160 - i['d'], i['i'], i['a'], file=out)
        elif i['e'] == 'PEERS':
            print(' <-', i['t'], 160 - i['d'], i['i'], i['a'], file=out)
        elif i['e'] == 'ABORTED':
            print('abort', file=out)
        elif i['e'] == 'COMPLETED':
            print('***', i['t'], 160 - i['d'], '\n', file=out)
        elif i['e'] == 'NEW':
            print('===', i['abstime'], i['type'], '===', file=out)
            print('<> ', 0, our_node_id, i['i'], file=out)
out.close()

out = open('dht_announce_distribution.dat', 'w+')
print('announce distribution items: %d' % len(announce_histogram))
for k, v in list(announce_histogram.items()):
    print('%d %d' % (k, v), file=out)
    print('%d %d' % (k, v))
out.close()

out = open('dht_node_uptime_cdf.txt', 'w+')
s = 0

total_uptime_nodes = 0
for k, v in list(node_uptime_histogram.items()):
    total_uptime_nodes += v

for k, v in sorted(node_uptime_histogram.items()):
    s += v
    print('%f %f' % (k / float(60), s / float(total_uptime_nodes)), file=out)
    print('%f %f' % (k / float(60), s / float(total_uptime_nodes)))
out.close()


print('clients by version')
client_version_histogram = sorted(list(client_version_histogram.items()), key=lambda x: x[1], reverse=True)
pp.pprint(client_version_histogram)

print('clients')
client_histogram = sorted(list(client_histogram.items()), key=lambda x: x[1], reverse=True)
pp.pprint(client_histogram)

out = open('dht.gnuplot', 'w+')
out.write('''
set term png size 1200,700 small
set output "dht_lookup_times_cdf.png"
set title "portion of lookups that have received at least one data response"
set ylabel "portion of lookups"
set xlabel "time from start of lookup (ms)"
set grid
plot "dht_lookup_times_cdf.txt" using 1:3 with lines title "time to first result", \
    "dht_lookup_times_cdf.txt" using 2:3 with lines title "time to last result"

set terminal postscript
set output "dht_lookup_times_cdf.ps"
replot

set term png size 1200,700 small
set xtics 100
set xrange [0:2000]
set output "dht_min_lookup_times_cdf.png"
plot "dht_lookup_times_cdf.txt" using 1:3 with lines title "time to first result"

set terminal postscript
set output "dht_min_lookup_times_cdf.ps"
replot

set term png size 1200,700 small
set output "dht_node_uptime_cdf.png"
set xrange [0:*]
set title "node up time"
set ylabel "portion of nodes being offline"
set xlabel "time from first seeing the node (minutes)"
set xtics 10
unset grid
plot  "dht_node_uptime_cdf.txt" using 1:2 title "nodes" with lines

set term png size 1200,700 small
set output "dht_announce_distribution.png"
set xrange [0:30]
set xtics 1
set title "bucket # announces are made against relative to target node-id"
set ylabel "# of announces"
set boxwidth 1
set xlabel "bit prefix of nodes in announces"
set style fill solid border -1 pattern 2
plot  "dht_announce_distribution.dat" using 1:2 title "announces" with boxes

set terminal postscript
set output "dht_announce_distribution.ps"
replot

set term png size 1200,700 small
set output "dht_lookup_distance_cdf.png"
set title "portion of lookups that have reached a certain distance in their lookups"
set ylabel "portion of lookups"
set xlabel "time from start of lookup (ms)"
set xrange [0:2000]
set xtics 100
set grid
plot ''')

dist = 0
for i in lookup_distance:
    if dist > 0:
        out.write(', ')
    out.write('"dht_lookup_distance_%d.txt" using 1:2 title "%d" with lines' % (dist, dist))
    dist += 1

out.close()

os.system('gnuplot dht.gnuplot')



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import glob
import copyright
import os


def update_file(name):
    if os.path.split(name)[1] in ['puff.cpp', 'puff.hpp', 'sha1.cpp', 'sha1.hpp', 'route.h']:
        return

    new_header = copyright.get_authors(name)
    subst = ''
    f = open(name)

    substitution_state = 0
    added = False
    found = False
    for line in f:
        if line.strip() == '/*':
            substitution_state += 1
        elif substitution_state == 1:
            if line.strip().lower().startswith('copyright'):
                # remove the existing copyright
                found = True
                existing_author = line.split(',')[-1]
                if existing_author in new_header or existing_author.strip() == 'Not Committed Yet':
                    continue
                print('preserving: %s' % line)
            elif not added and found:
                subst += new_header
                added = True
        elif line.strip() == '*/':
            substitution_state += 1

        subst += line

    f.close()
    open(name, 'w+').write(subst)


for i in glob.glob('src/*.cpp') + \
        glob.glob('include/libtorrent/*.hpp') + \
        glob.glob('include/libtorrent/aux_/*.hpp') + \
        glob.glob('include/libtorrent/extensions/*.hpp') + \
        glob.glob('include/libtorrent/kademlia/*.hpp') + \
        glob.glob('src/kademlia/*.cpp') + \
        glob.glob('examples/*.cpp') + \
        glob.glob('examples/*.hpp') + \
        glob.glob('tools/*.cpp') + \
        glob.glob('test/*.cpp') + \
        glob.glob('test/*.hpp') + \
        glob.glob('fuzzers/src/*.cpp') + \
        glob.glob('fuzzers/src/*.hpp'):
    update_file(i)



#!/usr/bin/env python3
from __future__ import print_function

f = open('../include/libtorrent/settings_pack.hpp')

out = open('settings.rst', 'w+')
all_names = set()


def print_field(str, width):
    return '%s%s' % (str, ' ' * (width - len(str)))


def render_section(names, description, type, default_values):
    max_name_len = max(len(max(names, key=len)), len('name'))
    max_type_len = max(len(type), len('type'))
    max_val_len = max(len(max(default_values, key=len)), len('default'))

    # add link targets for the rest of the manual to reference
    for n in names:
        print('.. _%s:\n' % n, file=out)
        for w in n.split('_'):
            all_names.add(w)

    if len(names) > 0:
        print('.. raw:: html\n', file=out)
        for n in names:
            print('\t<a name="%s"></a>' % n, file=out)
        print('', file=out)

    separator = '+-' + ('-' * max_name_len) + '-+-' + ('-' * max_type_len) + '-+-' + ('-' * max_val_len) + '-+'

    # build a table for the settings, their type and default value
    print(separator, file=out)
    print(
        '| %s | %s | %s |' %
        (print_field(
            'name', max_name_len), print_field(
            'type', max_type_len), print_field(
                'default', max_val_len)), file=out)
    print(separator.replace('-', '='), file=out)
    for i in range(len(names)):
        print(
            '| %s | %s | %s |' %
            (print_field(
                names[i], max_name_len), print_field(
                type, max_type_len), print_field(
                default_values[i], max_val_len)), file=out)
        print(separator, file=out)
    print(file=out)
    print(description, file=out)


mode = ''

# parse out default values for settings
f2 = open('../src/settings_pack.cpp')
def_map = {}
for line in f2:
    line = line.strip()
    if not line.startswith('SET(') \
        and not line.startswith('SET_NOPREV(') \
            and not line.startswith('DEPRECATED_SET('):
        continue

    line = line.split('(')[1].split(',')
    if line[1].strip()[0] == '"':
        default = ','.join(line[1:]).strip()[1:].split('"')[0].strip()
    else:
        default = line[1].strip()
    def_map[line[0]] = default
    print('%s = %s' % (line[0], default))

description = ''
names = []

for line in f:
    if 'enum string_types' in line:
        mode = 'string'
    if 'enum bool_types' in line:
        mode = 'bool'
    if 'enum int_types' in line:
        mode = 'int'
    if '#if TORRENT_ABI_VERSION == 1' in line:
        mode += 'skip'
    if '#if TORRENT_ABI_VERSION <= 2' in line:
        mode += 'skip'
    if '#endif' in line:
        mode = mode[0:-4]

    if mode == '':
        continue
    if mode[-4:] == 'skip':
        continue

    line = line.lstrip()

    if line == '' and len(names) > 0:
        if description == '':
            for n in names:
                print('WARNING: no description for "%s"' % n)
        elif description.strip() != 'hidden':
            default_values = []
            for n in names:
                default_values.append(def_map[n])
            render_section(names, description, mode, default_values)
        description = ''
        names = []

    if line.startswith('};'):
        mode = ''
        continue

    if line.startswith('//'):
        if line[2] == ' ':
            description += line[3:]
        else:
            description += line[2:]
        continue

    line = line.strip()
    if line.endswith(','):
        line = line[:-1]  # strip trailing comma
        if '=' in line:
            line = line.split('=')[0].strip()
        if line.endswith('_internal'):
            continue

        names.append(line)

dictionary = open('hunspell/settings.dic', 'w+')
for w in all_names:
    dictionary.write(w + '\n')
dictionary.close()
out.close()
f.close()



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import os
import re
import sys
from typing import Callable
from typing import Dict
from typing import Tuple

v = (int(sys.argv[1]), int(sys.argv[2]), int(sys.argv[3]), int(sys.argv[4]))


def format_fingerprint(version: Tuple[int, int, int, int]) -> str:
    ret = ""
    for i in version:
        if i < 10:
            ret += chr(ord("0") + i)
        else:
            ret += chr(ord("A") + i - 10)
    return ret


def fv(v: Tuple[int, int, int, int]) -> str:
    return f"{v[0]}.{v[1]}.{v[2]}.{v[3]}"


rev = os.popen("git log -1 --format=format:%h").read().strip()


def substitute_file(name: str, subs: Dict[str, Callable[[str], str]]) -> None:
    subst = ""
    with open(name) as f:
        for line in f:
            for match, sub in subs.items():
                if match in line:
                    line = sub(line)
            subst += line

    with open(name, "w+") as f:
        f.write(subst)


tab = "\t"
nl = "\n"

substitute_file(
    "include/libtorrent/version.hpp",
    {
        "constexpr int version_major = ":
            lambda ln: f"{tab}constexpr int version_major = {v[0]};{nl}",
        "constexpr int version_minor = ":
            lambda ln: f"{tab}constexpr int version_minor = {v[1]};{nl}",
        "constexpr int version_tiny = ":
            lambda ln: f"{tab}constexpr int version_tiny = {v[2]};{nl}",
        "constexpr std::uint64_t version_revision = ":
            lambda ln: f"{tab}constexpr std::uint64_t version_revision = 0x{rev};{nl}",
        "constexpr char const* version_str = ":
            lambda ln: f'{tab}constexpr char const* version_str = "{fv(v)}";{nl}',
        "#define LIBTORRENT_VERSION_MAJOR":
            lambda ln: f"#define LIBTORRENT_VERSION_MAJOR {v[0]}{nl}",
        "#define LIBTORRENT_VERSION_MINOR":
            lambda ln: f"#define LIBTORRENT_VERSION_MINOR {v[1]}{nl}",
        "#define LIBTORRENT_VERSION_TINY":
            lambda ln: f"#define LIBTORRENT_VERSION_TINY {v[2]}{nl}",
        "#define LIBTORRENT_VERSION ":
            lambda ln: f'#define LIBTORRENT_VERSION "{fv(v)}"{nl}',
        "#define LIBTORRENT_REVISION ":
            lambda ln: f'#define LIBTORRENT_REVISION "{rev}"{nl}',
    },
)

substitute_file(
    "Makefile",
    {
        "VERSION=": lambda ln: f"VERSION={v[0]}.{v[1]}.{v[2]}{nl}",
    },
)

substitute_file(
    "bindings/python/setup.cfg",
    {
        "version = ": lambda ln: f"version = {v[0]}.{v[1]}.{v[2]}{nl}",
    },
)
substitute_file(
    "src/settings_pack.cpp",
    {
        '"-LT': lambda ln: re.sub(
            '"-LT[0-9A-Za-z]{4}-"', f'"-LT{format_fingerprint(v)}-"', ln
        ),
    },
)
substitute_file(
    "test/test_settings_pack.cpp",
    {
        '"libtorrent/': lambda ln: re.sub(
            '"libtorrent/\\d+\\.\\d+\\.\\d+\\.\\d+"',
            f'"libtorrent/{v[0]}.{v[1]}.{v[2]}.{v[3]}"', ln
        ),
    },
)
substitute_file(
    "docs/header.rst",
    {
        ":Version: ": lambda ln: f":Version: {v[0]}.{v[1]}.{v[2]}{nl}",
    },
)
substitute_file(
    "docs/hunspell/libtorrent.dic",
    {
        "LT": lambda ln: re.sub(
            "LT[0-9A-Za-z]{4}", f"LT{format_fingerprint(v)}", ln),
    },
)
substitute_file(
    "Jamfile",
    {
        "VERSION = ": lambda ln: f"VERSION = {v[0]}.{v[1]}.{v[2]} ;{nl}",
    },
)



#!/usr/bin/env python3
# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4

import argparse
import os
import time
import shutil
import subprocess
import sys

from vmstat import capture_sample
from vmstat import plot_output
from vmstat import print_output_to_file

import platform

exe = ""

if platform.system() == "Windows":
    exe = ".exe"

def reset_download(save_path: str):
    rm_file_or_dir('.ses_state')
    rm_file_or_dir(save_path + '/.resume')
    rm_file_or_dir(save_path + '/cpu_benchmark')

def main():
    args = parse_args()

    ret = os.system('cd ../examples && b2 profile %s stage_client_test'
                    % args.toolset)
    if ret != 0:
        print('ERROR: build failed: %d' % ret)
        sys.exit(1)

    ret = os.system('cd ../examples && b2 release %s stage_connection_tester'
                    % args.toolset)
    if ret != 0:
        print('ERROR: build failed: %d' % ret)
        sys.exit(1)

    reset_download(args.save_path)

    if not os.path.exists('cpu_benchmark.torrent'):
        ret = subprocess.check_call([f'../examples/connection_tester{exe}', 'gen-torrent', '-s', '100000', '-n', '15', '-t', 'cpu_benchmark.torrent'])
        if ret != 0:
            print('ERROR: connection_tester failed: %d' % ret)
            sys.exit(1)

    rm_file_or_dir('t')

    run_test('download-write-through', 'upload', ['-1', '--disk_io_write_mode=write_through', '-s', args.save_path], args.download_peers)
    reset_download(args.save_path)
    run_test('download-full-cache', 'upload', ['-1', '--disk_io_write_mode=enable_os_cache', '-s', args.save_path], args.download_peers)
    run_test('upload', 'download', ['-G', '-e', '240', '-s', args.save_path], args.upload_peers)


def run_test(name, test_cmd, client_arg, num_peers):
    output_dir = 'logs_%s' % name

    rm_file_or_dir(output_dir)
    try:
        os.mkdir(output_dir)
    except Exception:
        pass

    port = (int(time.time()) % 50000) + 2000

    rm_file_or_dir('session_stats')
    rm_file_or_dir('session_stats_report')

    print("drop caches now. e.g. \"echo 1 | sudo tee /proc/sys/vm/drop_caches\"")
    input("Press Enter to continue...")

    start = time.monotonic()
    client_cmd = (f'../examples/client_test{exe} -k --listen_interfaces=127.0.0.1:{port} cpu_benchmark.torrent ' +
        f'--enable_dht=0 --enable_lsd=0 --enable_upnp=0 --enable_natpmp=0 ' +
        f' -O --allow_multiple_connections_per_ip=1 --connections_limit={num_peers*2} -T {num_peers*2} ' +
        f'-f {output_dir}/events.log --alert_mask=error,status,connect,performance_warning,storage,peer').split(' ') + client_arg

    test_cmd = f'../examples/connection_tester{exe} {test_cmd} -c {num_peers} -d 127.0.0.1 -p {port} -t cpu_benchmark.torrent'

    client_out = open('%s/client.out' % output_dir, 'w+')
    test_out = open('%s/test.out' % output_dir, 'w+')
    print(f"client_cmd: {' '.join(client_cmd)}")
    c = subprocess.Popen(client_cmd, stdout=client_out, stderr=client_out, stdin=subprocess.PIPE)
    time.sleep(2)
    print(f'test_cmd: "{test_cmd}"')
    t = subprocess.Popen(test_cmd.split(' '), stdout=test_out, stderr=test_out)

    out = {}
    while c.returncode is None:
        capture_sample(c.pid, start, out)
        time.sleep(0.1)
        c.poll()
    end = time.monotonic()

    stats_filename = f"{output_dir}/memory_stats.log"
    keys = print_output_to_file(out, stats_filename)
    plot_output(stats_filename, keys)

    t.wait()

    client_out.close()
    test_out.close()

    print(f'runtime {end-start:0.2f} seconds')
    print('analyzing profile...')
    os.system(f'gprof ../examples/client_test{exe} >%s/gprof.out' % output_dir)
    print('generating profile graph...')
    try:
        os.system('gprof2dot --strip <%s/gprof.out | dot -Tpng -o %s/cpu_profile.png' % (output_dir, output_dir))
    except Exception:
        print('please install gprof2dot and dot:\nsudo pip install gprof2dot\nsudo apt install graphviz')

    os.system('python3 parse_session_stats.py %s/events.log' % output_dir)

    try:
        shutil.move('session_stats_report', '%s/session_stats_report' % output_dir)
    except Exception:
        pass


def rm_file_or_dir(path):
    """ Attempt to remove file or directory at path
    """
    try:
        shutil.rmtree(path)
    except Exception:
        pass

    try:
        os.remove(path)
    except Exception:
        pass


def parse_args():
    p = argparse.ArgumentParser()
    p.add_argument('--toolset', default="")
    p.add_argument('--download-peers', default=50, help="Number of peers to use for download test")
    p.add_argument('--upload-peers', default=20, help="Number of peers to use for upload test")
    p.add_argument('--save-path', default=".", help="The directory to download to or upload from")

    return p.parse_args()


if __name__ == '__main__':
    main()
